/*
 * @(#)GTurtle.java   1.0 06/08/25
 */

// ************************************************************************
// * Copyright (c) 2006 by the Association for Computing Machinery        *
// *                                                                      *
// * The Java Task Force seeks to impose few restrictions on the use of   *
// * these packages so that users have as much freedom as possible to     *
// * use this software in constructive ways and can make the benefits of  *
// * that work available to others.  In view of the legal complexities    *
// * of software development, however, it is essential for the ACM to     *
// * maintain its copyright to guard against attempts by others to        *
// * claim ownership rights.  The full text of the JTF Software License   *
// * is available at the following URL:                                   *
// *                                                                      *
// *          http://www.acm.org/jtf/jtf-software-license.pdf             *
// *                                                                      *
// ************************************************************************

package acm.graphics;

import acm.program.*;
import acm.util.*;
import java.awt.*;
import java.util.*;

/* Class: GTurtle */
/**
 * The <code>GTurtle</code> class simulates a turtle moving on a canvas.
 * This class is similar to the <a href="GPen.html"><code>GPen</code></a>
 * class, but keeps track of both a position and a direction.
 */
public class GTurtle extends GObject {

/* Constructor: GTurtle() */
/**
 * Creates a new <code>GTurtle</code> object.
 *
 * @usage GTurtle turtle = new GTurtle();
 */
	public GTurtle() {
		setSize(TURTLE_IMAGE_SIZE);
		turtleVisible = true;
		penDown = false;
		direction = 0;
		animator = new Animator();
		setSpeed(DEFAULT_SPEED);
		erasePath();
	}

/* Constructor: GTurtle(x, y) */
/**
 * Creates a new <code>GPen</code> positioned at the point
 * (<code>x</code>, <code>y</code>).
 *
 * @usage GTurtle turtle = new GTurtle(x, y);
 * @param x The initial x coordinate of the turtle
 * @param y The initial y coordinate of the turtle
 */
	public GTurtle(double x, double y) {
		this();
		setLocation(x, y);
	}

/* Method: erasePath() */
/**
 * Erases the entire path drawn by the turtle but does not change its position.
 *
 * @usage turtle.erasePath();
 */
	public void erasePath() {
		path = new ArrayList();
		sx = 1.0;
		sy = 1.0;
	}

/* Method: setLocation(x, y) */
/**
 * Moves the turtle to the point (<code>x</code>, <code>y</code>) without drawing a line.
 *
 * @usage turtle.setLocation(x, y);
 * @param x The x-coordinate of the new position
 * @param y The y-coordinate of the new position
 */
	public void setLocation(double x, double y) {
		super.setLocation(x, y);
		animator.delay();
	}

/* Method: move(dx, dy) */
/**
 * Moves the turtle by the specified displacements.  The turtle draws a line if its
 * pen is down, but otherwise simply moves.  This definition is included because it
 * exists in the <code>GObject</code> superclass but will usually be avoided in
 * favor of <code>forward</code>.
 *
 * @usage turtle.move(dx, dy);
 * @param dx The distance to move the object in the x direction (positive is rightward)
 * @param dy The distance to move the object in the y direction (positive is downward)
 * @noshow
 */
	public void move(double dx, double dy) {
		if (penDown) {
			double x = getX();
			double y = getY();
			synchronized (path) {
				path.add(new SetLocationElement(x, y));
				path.add(new DrawLineElement(dx, dy));
			}
		}
		super.move(dx, dy);
		animator.delay();
	}

/* Method: forward(distance) */
/**
 * Moves the turtle forward in its current direction by the specified distance.
 * The turtle draws a line if its pen is down, but otherwise simply moves.
 *
 * @usage turtle.forward(distance);
 * @param distance The distance to move in the current direction
 */
	public void forward(double distance) {
		if (penDown) {
			double x = getX();
			double y = getY();
			synchronized (path) {
				path.add(new SetLocationElement(x, y));
				path.add(new DrawLineElement(distance * GMath.cosDegrees(direction),
				                                    -distance * GMath.sinDegrees(direction)));
			}
		}
		movePolar(distance, direction);
		animator.delay();
	}

/* Method: forward() */
/**
 * Moves the turtle forward by its own size.
 *
 * @usage turtle.forward();
 */
	public final void forward() {
		forward(getTurtleSize());
	}

/* Method: setDirection(dir) */
/**
 * Sets the direction in which the turtle is moving.
 *
 * @usage turtle.setDirection(dir);
 * @param dir The direction in which the turtle is moving
 */
	public void setDirection(double dir) {
		direction = dir;
	}

/* Method: getDirection() */
/**
 * Returns the direction specified by <code><a href="#setDirection(double)">setDirection</a></code>.
 *
 * @usage double dir = turtle.getDirection();
 * @return The direction in which the turtle is moving
 */
	public double getDirection() {
		return direction;
	}

/* Method: left(angle) */
/**
 * Changes the direction of the turtle so that it moves the specified number
 * of degrees to the left.
 *
 * @usage turtle.left(angle);
 * @param angle The angle to turn, measured in degrees
 */
	public final void left(double angle) {
		turn(angle);
	}

/* Method: right(angle) */
/**
 * Changes the direction of the turtle so that it moves the specified number
 * of degrees to the right.
 *
 * @usage turtle.right(angle);
 * @param angle The angle to turn, measured in degrees
 */
	public final void right(double angle) {
		turn(-angle);
	}

/* Method: isPenDown() */
/**
 * Returns whether the pen is down, as described in
 * <a href="#setPenDown(boolean)"><code>setPenDown</code></a>.
 *
 * @usage if (turtle.isPenDown()) . . .
 * @return <code>true</code> if the pen is down, <code>false</code> otherwise
 */
	public boolean isPenDown() {
		return penDown;
	}

/* Method: penDown() */
/**
 * Lowers the pen.
 *
 * @usage turtle.penDown();
 */
	public final void penDown() {
		penDown = true;
	}

/* Method: penUp() */
/**
 * Raises the pen.
 *
 * @usage turtle.penUp();
 */
	public final void penUp() {
		penDown = false;
	}

/* Method: setColor(color) */
/**
 * Sets the color of the turtle's pen.
 *
 * @usage turtle.setColor(color);
 * @param color The new color for the pen
 */
	public void setColor(Color color) {
		synchronized (path) {
			path.add(new SetColorElement(color));
		}
		super.setColor(color);
	}

/* Method: isTurtleVisible() */
/**
 * Returns whether the turtle is visible.
 *
 * @usage if (turtle.isTurtleVisible()) . . .
 * @return <code>true</code> if the turtle is visible, otherwise <code>false</code>
 */
	public boolean isTurtleVisible() {
		return turtleVisible;
	}

/* Method: showTurtle() */
/**
 * Makes the turtle visible.
 *
 * @usage turtle.showTurtle();
 */
	public final void showTurtle() {
		setTurtleVisible(true);
	}

/* Method: hideTurtle() */
/**
 * Hides the turtle.
 *
 * @usage turtle.hideTurtle();
 */
	public final void hideTurtle() {
		setTurtleVisible(false);
	}

/* Method: setSpeed(speed) */
/**
 * Sets the speed of the turtle, which must be a number between 0 (slowest)
 * and 1 (fastest).  Setting speed to a value less than one makes the turtle
 * move slowly, thereby making it easy to see exactly how a figure is being drawn.
 *
 * @usage turtle.setSpeed(speed);
 * @param speed The speed of the turtle (0 is slowest, 1 is fastest)
 */
	public void setSpeed(double speed) {
		animator.setSpeed(speed);
	}

/* Method: getSpeed() */
/**
 * Returns the current speed of the turtle.
 *
 * @usage double speed = turtle.getSpeed();
 * @return The current speed of the turtle (0 is slowest, 1 is fastest)
 */
	public double getSpeed() {
		return animator.getSpeed();
	}

/* Method: setTurtleSize(size) */
/**
 * Sets the size of the turtle.
 *
 * @usage turtle.setSize(size);
 * @param The size of the turtle image in pixels
 */
	public void setSize(int size) {
		turtleSize = size;
	}

/* Method: getTurtleSize() */
/**
 * Returns the size of the turtle.
 *
 * @usage int size = turtle.getSize();
 * @return The size of the turtle image in pixels
 */
	public int getTurtleSize() {
		return turtleSize;
	}

/* Method: paint(g) */
/**
 * Implements the <code>paint</code> operation for this graphical object.  This method
 * is not called directly by clients.
 * @noshow
 */
	public void paint(Graphics g) {
		PathState state = new PathState();
		state.sx = 1;
		state.sy = 1;
		synchronized (path) {
			int nElements = path.size();
			for (int i = 0; i < nElements; i++) {
				PathElement element = (PathElement) path.get(i);
				element.paint(g, state);
			}
		}
		finalElement.paint(g, state);
		if (turtleVisible) drawTurtle(g);
	}

/* Method: getBounds() */
/**
 * Returns the bounding box for the entire figure traced by the turtle.
 *
 * @usage bounds = getBounds();
 * @return A <code>GRectangle</code> representing the bounding box
 */
	public GRectangle getBounds() {
		PathState state = new PathState();
		GRectangle bounds = new GRectangle(-1, -1, -1, -1);
		state.sx = sx;
		state.sy = sy;
		synchronized (path) {
			int nElements = path.size();
			for (int i = 0; i < nElements; i++) {
				PathElement element = (PathElement) path.get(i);
				element.updateBounds(bounds, state);
			}
		}
		return bounds;
	}

/* Method: contains(x, y) */
/**
 * Contains is defined to be false for the <code>GPen</code> object to avoid having the
 * trace intercept mouse clicks.
 *
 * @noshow
 */
	public boolean contains(double x, double y) {
		return false;
	}

/* Static method: main(args) */
/**
 * Makes it possible for a <code>GTurtle</code> to be run under the control
 * of a <code>GraphicsProgram</code> object.
 *
 * @usage main(args);
 * @param args An array of string arguments
 */
	public static void main(String[] args) {
		String[] newArgs = new String[args.length + 1];
		for (int i = 0; i < args.length; i++) {
			newArgs[i] = args[i];
		}
		newArgs[args.length] = "program=acm.program.GObjectProgram";
		Program.main(newArgs);
	}

/* Inherited method: setLocation(pt) */
/**
 * @inherited GObject#void setLocation(GPoint pt)
 * Sets the location of this object to the specified point.
 */

/* Inherited method: getLocation() */
/**
 * @inherited GObject#GPoint getLocation()
 * Returns the location of this object as a <code>GPoint</code>.
 */

/* Inherited method: getX() */
/**
 * @inherited GObject#double getX()
 * Returns the x-coordinate of the object.
 */

/* Inherited method: getY() */
/**
 * @inherited GObject#double getY()
 * Returns the y-coordinate of the object.
 */

/* Inherited method: getX() */
/**
 * @inherited GObject#double getX()
 * Returns the x-coordinate of the object.
 */

/* Inherited method: getY() */
/**
 * @inherited GObject#double getY()
 * Returns the y-coordinate of the object.
 */

/* Inherited method: getSize() */
/**
 * @inherited GObject#GDimension getSize()
 * Returns the size of the bounding box for this object.
 */

/* Inherited method: sendToFront() */
/**
 * @inherited GObject#void sendToFront()
 * Moves this object to the front of the display in the <i>z</i> dimension.
 */

/* Inherited method: sendToBack() */
/**
 * @inherited GObject#void sendToBack()
 * Moves this object to the back of the display in the <i>z</i> dimension.
 */

/* Inherited method: sendForward() */
/**
 * @inherited GObject#void sendForward()
 * Moves this object one step toward the front in the <i>z</i> dimension.
 */

/* Inherited method: sendBackward() */
/**
 * @inherited GObject#void sendBackward()
 * Moves this object one step toward the back in the <i>z</i> dimension.
 */

/* Inherited method: getColor() */
/**
 * @inherited GObject#Color getColor()
 * Returns the color used to display this object.
 */

/* Inherited method: setVisible(visible) */
/**
 * @inherited GObject#void setVisible(boolean visible)
 * Sets whether the trace of the turtle is visible.
 */

/* Inherited method: isVisible() */
/**
 * @inherited GObject#boolean isVisible()
 * Checks to see whether the trace of the turtle is visible.
 */

/* Protected method: start() */
/**
 * Starts a <code>GraphicsProgram</code> containing this object.
 *
 * @usage gobj.start();
 * @noshow
 */
	protected void start() {
		start(null);
	}

/* Protected method: start(args) */
/**
 * Starts a <code>GraphicsProgram</code> containing this object, passing
 * it the specified arguments.
 *
 * @usage gobj.start();
 * @param args The array of arguments
 * @noshow
 */
	protected void start(String[] args) {
		GraphicsProgram.startGraphicsProgram(this, args);
	}

/* Protected method: drawTurtle(g) */
/**
 * This method draws a representation of the turtle at the current location.
 * Subclasses can override this method to draw fancier turtles.
 *
 * @usage drawTurtle(g);
 * @param g The graphics context in which to draw the turtle
 */
	protected void drawTurtle(Graphics g) {
		Component comp = getComponent();
		if (comp == null) return;
		Image image = createTurtleImage(direction);
		int imageSize = image.getWidth(comp);
		int x = (int) Math.round(getX());
		int y = (int) Math.round(getY());
		if (imageSize == turtleSize) {
			g.drawImage(image, x - imageSize / 2, y - imageSize / 2, comp);
		} else {
			g.drawImage(image, x - turtleSize / 2, y - turtleSize / 2, turtleSize, turtleSize, comp);
		}
	}

/* Protected method: getTurtleBounds() */
/**
 * This method returns the bounds that the turtle occupies.
 *
 * @usage Rectangle r = getTurtleBounds();
 */
	protected Rectangle getTurtleBounds() {
		Component comp = getComponent();
		if (comp == null) return new Rectangle();
		Image image = createTurtleImage(direction);
		int width = image.getWidth(comp);
		int height = image.getHeight(comp);
		int x = (int) Math.round(getX());
		int y = (int) Math.round(getY());
		return new Rectangle(x - width / 2, y - height / 2, width, height);
	}

/* Factory method: createTurtleImage(dir) */
/**
 * This method creates (or retrieves) an image of the turtle in the specified
 * direction.  Subclasses can change this method to create different images, but
 * the images must be square.
 *
 * @usage Image image = createTurtleImage(dir);
 * @param dir The direction in which the turtle is heading
 * @return The image used to draw the turtle
 */
	protected Image createTurtleImage(double dir) {
		int heading = GMath.round(dir / 10);
		if (heading < 0) heading = 36 - (-heading % 36);
		heading %= 36;
		if (turtleImage[heading] == null) {
			turtleImage[heading] = MediaTools.createImage(TURTLE_IMAGE[heading]);
		}
		return turtleImage[heading];
	}

/* Private method: turn(angle) */
/**
 * Changes the direction in which the turtle is moving by rotating it <code>angle</code>
 * degrees counterclockwise.
 *
 * @usage turn(angle);
 * @param The angle to turn, measured in degrees
 */
	private void turn(double angle) {
		direction += angle;
		repaint();
	}

/* Private method: setTurtleVisible(visible) */
/**
 * Sets whether the turtle itself is visible.  If the turtle is visible, the <code>drawTurtle</code>
 * method will be called at the end of painting the path to show the current location.
 *
 * @usage turtle.setTurtleVisible(visible);
 * @param visible <code>true</code> to show the turtle, <code>false</code> to hide it
 */
	private void setTurtleVisible(boolean visible) {
		turtleVisible = visible;
		repaint();
		animator.delay();
	}

/* Private constants */

	private static final double DEFAULT_SPEED = 0.5;

/* Private instance variables */

	private PathElement finalElement = new FinalPathElement();
	private Animator animator;
	private double sx, sy;
	private double turtleSpeed;
	private double direction;
	private boolean penDown;
	private boolean turtleVisible;
	private int turtleSize;
	private int delayCount;
	private ArrayList path;

/* Turtle images */

	private static final int TURTLE_IMAGE_SIZE = 68;

	private static Image[] turtleImage = new Image[36];

	private static final String[] TURTLE_00 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE0000002200003200000000C30000C30000C30000C30000C30000C30000C30000C3",
		"0000C3000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C3A9F50",
		"26C5A548143E9F42A311DD4AAA8D07686B8B3A6DB66D438B22A9B7A40269FE5E6F296A371A6D2EDA",
		"1008247B693614786B0956050101200958201F22108F8C8C1F018A115A700D3604147D46680D8788",
		"05090D018B979699AB8C19AC8B725B3114853624098D201B05654B6120060FB978B8C877CA9ECD84",
		"0D28692428052E9E36167CA2DBDC42368B0FE1DADDE4E5E6E7E84B85B8DB36E20D35E94286355B2E",
		"B8681687EBEB142486A5F00822344EDE214369E4CC51B0A68001588958F16AE02001094673FA15E4",
		"D62086344F78221628902141B84C1052524A04E20104478C12A9C9B58684015E1F1E2850106C490D",
		"FF2D5340D6C933D2D2485546676542B18C10471B1830F44C32C8068CAB5607791C54C3860B182E48",
		"B46C29559ED9B368D3AA5D6B161781B4261AEC04616DEA93060850396A26CFAB1B031515180051B5",
		"30D73D734884F574164D0D17D86C7834B46F269966FBD294BA865617B26761B514CD34D2CA2A1215",
		"B658D4437963943D44345B7BA3C5C1482C3131BD5AC50A16273203B9E90204D9CDA1A2A978997C90",
		"008402959112B914016251751723073598C392611F819A6B6E19193DC007102941C03409E1BCDEDF",
		"7812EACD0941DCEB349FE1A4DAEEDB34AAFF123D74071A75D9D5852373DC4306324C0D751453BF35",
		"68077EF8B0C31102835D634807F818A6FF86294C31151E415EE9734E2E691532472F2D3D00035B30",
		"C628E38C34D668E38D38E6A8E38E333650C103A1A8E589020800294A032054649F596834A241001A",
		"7C104F172EB8D05C010A2C898E5C0A6C9080061BE464A0120D545926960C7565D66617B810432983",
		"E0A26140FC1416DF98E614E2020393DD22E77682DCD10C6D025DE31A3AB98477591A12DAF19B509A",
		"B9C15963F85936DE8A015CC44B6905540021196820C405A203D5818A0271ACE1DB68A824609B2597",
		"C16901A27436B80B43B06492812BB9FAC68802BFC1260A369095B187862389561A7A38A987852CE6",
		"F1028B1D9759D887570848A0182983A2E2062BE831128E4BED3DF21C0801409BFFE982C7C2E3C994",
		"67F858AC036AC8818AAA01A4B4AB4B26982042262DF9AB1BBA6FE0A300060AA897C04E7852A59929",
		"47D99B6F161F34A712C0CFFDCBD2519625ACF0C23BC1EBC49D85A896CA6D582C0249CA9754CC2B27",
		"68C06041048AF082AA20F7691629091100ABCA22CA25421AD0406FA7051A196555D813921D24CD9F",
		"F55EA41B2D9B320234697A90722814362C4C4795485B0AA80374AC1101CFDF729775850D338DAE15",
		"36B4592C62F8E4F18C40920E182A3FE4D4C05C230D1050419555C627AA9C07C9F90F19B6226DAD70",
		"558C21AC1F7BE0D3952E9EA5562C059707A5183C0F1FE94EDBDDE82917CD1180DACE8B29E2358B53",
		"3C9209C320B1D76E07FBEDB8E77E56100021FF0B4D414347436F6E20040310390000000157726974",
		"74656E20627920474946436F6E76657274657220322E342E33206F66204D6F6E6461792C204D6179",
		"2032352C2031393938003B"
	};

	private static final String[] TURTLE_01 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE000000220000320000000078000078000078000078000078000078000078000078",
		"000078000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C3A871C",
		"15EAF2AC1639A39142451A55152357C35548911A366B95C21EC16C9498804D69A0EDF6023EEFA0D1",
		"1A7A0D27730D033631136A4A08761402367F80760E05092220011F1F202222091F01190A1020A005",
		"2005628D36362412060999200A0A354D2C2A1C0D31368D78056780059A1F05A1190101B105C609CD",
		"CC76AC272736298A45287A1467238F03767E76277B82D3D37B2728E3D4362314232D0D3524F4696A",
		"2A0D121815D6D7FEFF00030A1C48B0A0120505227C2A7082820A831D5C6851904083860F1A22A460",
		"C5D186804674F6D43933084D358842586C1C5747D01812818CC964962C19CD9B7B7A391A8180C390",
		"FF142C9AD8B090C2A4B471BD00296B402219080821323C85402AC127109C3E58D59AF3458C91B112",
		"90A87042D1461BEBC6D98109A829A800D09219D0034C924E01020C095440028481072708D47041D8",
		"5147568620BD786183718AC748588C6860A692581227516ADECCB9B3E7CF4CCC48E8650B7494A05F",
		"C29070062281A07E10539CD210404381D247549E3D2CCD246C8329B651A8500186510B2D1B7C2C24",
		"A857D28677989A7C94ABE046140304A825B776E64CB8F428D5AC57E7913D827696E3A1879066820C",
		"C86ADACC549319343BD959ADB57AF97C1216A875C0025ACAC1218C33C604F094535491A2D5070F98",
		"20C25B5691338D65C46075C66F46A4D0FF420B292010DC7679C0649F8222C0B509555825A8492C70",
		"81374D3D3630C0E1352D21660782269E58DF8934D5D44721780D4041411C0CD20B09149058A2036C",
		"A12009776B8D308D23372AC2C507F4B8D3820B144863647A8524B58794358CC18A0028D850830D2D",
		"BCA380421AD0F30005013A51C3031114B36696020215500344097A0D0770F807DAA28C36EAE8A390",
		"462AE9A49406A400057756EA440311607567759A1EC1174556817027A004A520400B233C54C51912",
		"20448202F21CB9D9A820C886C2131234D40A9393D990A741ED28A3093023A0A65B986B7AE4C8206A",
		"A1C511AA008D4A4204942450810368E035CD0086D4F10D77FA3530996F103540FF410D7580E9D123",
		"E84443A52F78AC6918B56AD860E405792D96D676CCD0639F7D15E4816039EA866B1D81F3E691205C",
		"411AD3800362258350BD98290A50511F852948257CE961408CF225F8B07D545223000935D4700209",
		"29F864C53B576AD79658347D101F313ACB07DF7C0104D3912B0A6095C02C0A68AC440A2C9D9B8730",
		"CAB4A689A7543DF80004131A83091EBE327555A9A6C29C042B1404D7C15163022296520A3E70F453",
		"2648F8E0531326A35520E29C50C1C015454099123688181C092ED8BC074E4E3D00CA8A55478DF55B",
		"F5D9F195472F18F5D8B091B1DA40051E37E7705BC36815CA07971C9B60060932A5072B62C0319047",
		"F2E2516190762763BE954D3E422CCE23D3E05B850AE71C4E4F0474DD943BC410AF9E68A20BEB5426",
		"393EC6140109DC4AC935D9FA1A82B92271D4116C72B11F9E37390CD7FBCE000808A1820208A4AF06",
		"0BABD1D3451DD44CD3CB08D951498121CC99D9261C85A387A9EA94B17B9000140968081BD4B54038",
		"18821A76A8817630430201B88C17D510942346123AB1F4CE0A7B8A0508D2C1860E944D836CE80805",
		"18C08027E06232C0A04709ADC00505700966BEA354141A72A9503D61283E0CA2108748C421060100",
		"21FF0B4D414347436F6E2004031039000000015772697474656E20627920474946436F6E76657274",
		"657220322E342E33206F66204D6F6E6461792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_02 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE0000002200003200000000DEAB00DE0000DE0000DE0000DE0000DE0000DE0000DE",
		"0000DE000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C365D92",
		"1449E2AC5A9D0AD7A841228D6CB6D4D58A6829129A48244172D85E8286DC3697DB1B854285D64814",
		"0272277127843614452C122C63422914740D71725D24790119010109202022201FA01F0F10221F99",
		"0909770D821411240901240D614B292F0323278F02607779957979A07E011F98A7C17FB09A36BA60",
		"032D8B8C49BD140D1427AAAA2474DD7627036028601286235DCB090F2487571C29201A1A09348608",
		"D3F8F9FAFBFCFDFEFF5D0A7C28A04056BB7F45B8814821471A111629C0F0A260CD4EB86C5C66D9E0",
		"D5C0213F041406506000231C0C6D052A650A106C65A63F2C8345283087970D121836C042A5008402",
		"FF1B64E450B4930CA628109B4641800042D30710104CADCCA3D1C5AC7D143878693607639F02AE68",
		"B27C19CC409706282C9E6047670405170A2ADC2C97C263951A29BD6C69106E235F8A60BAA17841B8",
		"2EC226B32A3CB061F7B0E3C790F571887CD8C6064E0AB029A0AC8F4405096CCE1E8CCC424548154D",
		"1C8DEB05E644E0308DFB358D37905B0747AF05885C556756C54060863E7AE1CF86D50A2E44BE19A0",
		"DB4E5A2E9560FDC9E3A700B1587720B9E6C325C107050ABE5479848263F67334F118B0B43241863F",
		"984CBDCF242B50B8060A407D422A4B4C1586AC4D4215759F18F380529F780755299A6C925E03B76C",
		"F48205B159D11A384469824726103C80FFCA53A47C02D325CAA4E79A4D8C3142810A2DA490820A29",
		"7C818203B3C841958D236A32D67431E54553338F3C02D43FD61812881DF515250B0DDCD9D18C0D28",
		"6403460B149110C1066A90F0C04FD36875CD2CBC1894D62E77FCB6D60035D860817F882CC2100C86",
		"31C242034DF921000AB8B8909C214106C7266748D49042042074F127A0FA94A626A28C36EAE8A390",
		"462AE93E201130D9A45668B5C14D4362CA84561FB29199A74B8C90022A9B3880C2A1A40EC1410315",
		"90E042015274D4280B152251CF93521A02280B24ACA080004AB0206438CD7D33C01C86B0CA8F560A",
		"C8F38103241011915F148DB0D71DAE4D328B48BC8C70293F360CC0522820B8B2FFD120155504094A",
		"19757B0D18B992418202331950035FCF2C7B245138B25181030EC094D28374847300050828B02F25",
		"CE22C18E4521C52000057140A25251A700E2404C2C692324749C6C02DE674E94FB56B2DAE0C1A326",
		"3CB287CC1FB2D024081D60698014548576CA040BBD7C59E32BC02813DF25C6ACD489297F98F24D8D",
		"C92848C2095740940B515D28E0F253A684B25454A83C6002830D22FCC62373C479850D5B84644393",
		"2E67B2F426509900B6C9108430D054720C208BAFFB40F2C8B2D46DA8C952987CBDD432A08C058841",
		"ACD1320644B842C495942E5BB761069C64F294278DCBACC97A1619E2B3852E20C08EAC71D078871F",
		"9917980926A8BC6CC230208484A4DB6855A790CD0815A08D121B6235B063327C9B5D08C66C16E402",
		"3E1059E3DABB94D094924B38B69C9D0D2D8C306B4F05E912716A8109F52ED604A7E78DD072EC8A02",
		"05DAE26740824FD6BB04365CA9829157BC5D7342935F308434AC061DEBB0657C4A28D735BAE78267",
		"4CCF0E71C0861CC2151804126133F960C1192831800B740F63454291006C42A148D9A0291B20C1C5",
		"2A9082150C406DB721150714E31D8DA4A8554850C12B362001DEE1F0085AC198002CF8C3221AF188",
		"FE08020021FF0B4D414347436F6E2004031039000000015772697474656E20627920474946436F6E",
		"76657274657220322E342E33206F66204D6F6E6461792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_03 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE000000220000320000000022AB0022000022000022000022000022000022000022",
		"000022000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C223995",
		"548396E238AF5827A7E52291280D9B384B162A6CA054A1401AA6C4B6810D7CB235EE8D4285960FF8",
		"4905780D27246176120A1F011F0A200F24364D1C14940C313103772877236B8106011901091F2022",
		"1FA809090F262209A3A46B770385362965462D279C717705AA7D09A6AFA88AA301BFC80501240986",
		"620314912C738F622C572C292C1C2A14996177768278857876D19B603616D8B8D9CD5E0D0328142D",
		"2DD1EEEFFB4C1C1C2323D8F11B48B0A0412C1CD47C280082048D480709BE1120E0049875BB1A90A8",
		"4369C08017108D48E086250505051A12188840428E8D18784E6C5A4342D9323FCA7026B81948DC20",
		"FF3BBF9641BA45C664343C6CE43958564A110810104C4038F56A518635C802DCD925604EC82CD4BA",
		"8A1110D05C0301BECCCACAAA2C4F843BB5EAA0B0215700178D581F9CA1C06FC480AE720AC9014717",
		"8E8D17ED225E69A02081864214F4299E4CB9F21515562C47B4B12102880223CE681ED8C045E3040A",
		"068DDEC7A282981375A0897BDD66320B172D462038C2425A344A787C0B3217675D401B2A5C74D108",
		"42817312159C3C60A8C1294BD98332E6F1B52CE9D29C9E7A1A66FEF4832A679291A44E017D0405BF",
		"31E66867D3DD4FD5AA37458030A64AD6BCC32F7C145919D4D8501139BBFC024801C590F2D4035329",
		"228C0211FAD1DD33BBF045DA1DD26CB2FF9D277E4025C228A544A8085638D9379747D17CD5C16E65",
		"48534738793443D32FA8D86753556669E2402D68790142029E81B04E7A45B0F0020229A8200102B9",
		"8D0386203BB1546377C49DC549187574750B366F2840412144F1238D1D1765A4269AC111224699AB",
		"A500D37B0035701C6186D9B2DA1227BCF25803F5C0B8272E5E908081688326AAE8A28C36EAE8A346",
		"FC43810B4842AA84028890F0801D955A4A040743AAF2DC089D7ADA01095D244002085394EA290770",
		"79A18B0DAE5ADA808C623C3307057042DA1B1CE4D8F9538B36B430016508B858C444BB0A92494670",
		"A075AB189410A100215ED8B042ADCBA6B002244648831625D00AC287AAC89030FFC239D0DAD14C79",
		"7A89D96B122928B041001ADCC14281684E394E4F4955B0544D1692439118CE1920CC2A10A83BEF11",
		"1478A2C1060A1810C6511B9553537F5981689350FEC98197274F85C1EDA40ABC55004CD41E350E20",
		"F685871583A48812A256E4B05381467364F15E03359C65200CDA7137738AA888920185229CE2C742",
		"3D35F002051CBD830F250109B24BBD0B22FD41D3119202210427E267880B613C9C4D8119FE2BCF1A",
		"C2EC9781235199605EDC0F9857F06C5D35A0B6CF59C386541E37399240061F40A0380443A6B8934A",
		"0538C0E1582A0CC482B4DF70074C320140F0C0D3A68090C12BE15918089A714C3D446ABC726BC438",
		"2778A479E9398602CC4B77B89B655118235860C1080E34E285022E68B804981DBCC6613951A3D831",
		"EE069020391F39DB8575C5FB25A56C1114C4D0027B15A07AA0E48204A56A2005646073B0456B6243",
		"051408100D096A144089EB29A480807630D06567AEC3411F4DB6C325916DA2055E1A0D0206A09DBF",
		"8C235810FCDF8006552008AAE328223384EA1885B616D8492CB9FAC699BC422B478121292E18016E",
		"6EE50239F02A32C67BD49840308AD811200AFE30156F5052C337E9100912A800F626F84324B0A00E",
		"0378C3DF8A48842532D15241000021FF0B4D414347436F6E2004031039000000015772697474656E",
		"20627920474946436F6E76657274657220322E342E33206F66204D6F6E6461792C204D6179203235",
		"2C2031393938003B"
	};

	private static final String[] TURTLE_04 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE000000220000320000000018AB0018000018000018000018000018000018000018",
		"000018000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C125D14",
		"4583C6E238AFD82BA7511991280D9B9842A6A814CB6A164BB10DC88D70BCE0A8340A8580BE70E7C7",
		"6D6D270D2478790D146C0D098B0E0D03636F71837801947A1F1F099909200F1010987B950D026E02",
		"6F6B472C2D24823628927D797A0F201F0198B78B099501B1BF80508384015F884D29362D62232EA5",
		"36248EB10524D592D1BF92A52CA9DD43168723642D2D6F2E80DEE94E242924180F64EAF2F3F4F547",
		"1C0ABB8B35C7F6F66D6EFED89074C28D200A05473C42018B04275C0A6AB1BBC262848214091C90E8",
		"C0C28680820304C521C46B56C90009326408B01253023F61DE88E1966AC4A11A0C46BCC8E6504121",
		"FF10223E64F8A00082894F1078E959CAB4C1090A8F1E8DE820A18102290AA064011367C0A33E0EE7",
		"14CA336B541CAE609E51A0E9AF0807310329608B83E284CD678F6CA4689B4551A60705E0F21D4CB8",
		"B05BC36D2382006195840DB688536D71F15241B47E915349816BA320578F50C5BCD84B84C5201057",
		"2D57E0DB51001C495CA945B326091A4012963FD8BAB5A10066261C28A4180099054001AEE7CCA994",
		"8084835985F4CC1569C302F2C7A95228D0902B4204375C1BC032894729CB59202004D5F37054E765",
		"E8B204BF48A202390A02628C148BC7C0524C9DA8A7CB620F64924749B5C1C5D61623B00059120085",
		"F715352F3D97876E20B4B41804B69414FF5D317C0012C608AC9000C2222014A300764DAC721618B1",
		"2DC2C72CB9B0F71225CE0D12464153A0F0580A66D8C0410A1C58919D18228131C01C0E15D0DC4864",
		"11F207057571519D3FCAA4D04E0A71C03090535FF2F8476D8E7C69C30B16648644326DE01590479C",
		"3DA866690E58160D332D0060E49C5A90F08026099C20279F585831240BA411AAE8A28C36EA680723",
		"10C0D1A3575453CD8A0D0C4A29113678879A14536D9A043B95DD292A125B48E05335339D6A04072D",
		"8C30028F487226860A0C20C1950BE72056019224D8B41F6887E4158F0B080C82226A20C4D3967162",
		"BCC18748C00830A64056E9D3C965FEB4F6542C82D070875261C502A718052860FF202787689AC45B",
		"7B12D146195F4A8247050E3870A321B2E1E1541B714460CD6F4EE825C56A427474C2094B3A75821F",
		"BD9065094A78DCF24B411EBDE0EE1283DC424D14D10E22EC30638952C02DBA8582D24BB17811DF1A",
		"2638A9C906CD891143BD2389629E501968929E0927BEE461189D0D9428162D6817816F6FC0508A1C",
		"D3F8A7A1089F04E5735199ECE1C74051B57018130A0CD0C25930BCB093784EB960D91DB864A29B27",
		"46D9F29F4A07B20CDA297A256CD5232F30010505FCE0CCA47315039532DC1C82309625D4105D4A19",
		"1438F0A789771EAD0AB4624C834D749CA88752CAB7885249E38788EC9123B845A00B3C042B311020",
		"AFF4B14834BCF40CC24A25BA1CA887C0B0C441434105A1E01A05B2525080C66B94731672FCED7B72",
		"E8A3FBEBEF5C7130CC623D1D75462D7FD334C0D24AD80CC4A3B80DC4D0BA3C6BC13E17CE2F85450D",
		"358DC81149056D582E4F0A68AA9042D20838136667D908E021424301E4B1661CCBC08783C080B3AF",
		"38E510C9094918D0D4A8EC894958000C8D8228158D1650606C445B1220D6B2B139E1671850100739",
		"0450C245B180049B981501C8102F5771840229AA44AB6C58840AE08613A8E0A11158B0B0111A5088",
		"8E0A020021FF0B4D414347436F6E2004031039000000015772697474656E20627920474946436F6E",
		"76657274657220322E342E33206F66204D6F6E6461792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_05 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE0000002200003200000000B70000B70000B70000B70000B70000B70000B70000B7",
		"0000B7000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C762E1C",
		"96734AAD0A152D498A446AD8BE6030652025561A89CF07047A90525627C546A134EE785AA3400AF8",
		"0B05115C7877272F5F7071422E0A2909050D297336772494807E7E097E6A6A09096D20101F9B019B",
		"053427021402032F70151420240A0A5E65470A7403352E2394840D248F7B99A20F9B6A9A01057E11",
		"78C00D0336B88A4873787627767705095C5DC17776362F16D6E8442A0D14232DEC03032DD4E9F54A",
		"0DB506201275F6FEFF00ADA4D8100104A85B01137640F1E5CE08425F4ED81870C7C6094B5F481450",
		"40AA19090A015980894708932966DE028050666A25332F0DD92932D1AC8006071507AC3AD10D50B3",
		"FF00A4D43C8000C10451833E7D06E05280120989ED620853A1114635272C469C18806D0FA6474951",
		"32CB54B1228D313614B26A41224B8A163180456397E7A2B430571536A9306C980D02040624D24BB8",
		"B0E1C38A5C5852F0B11F62744CBFD93A9177090570B62A18B6E5A28144895C61D6616751950D86B4",
		"3EA521A5E1046191132940C2C307D0A060BB8009B254D9DE9CD1C1F84428E02093B7660635E27931",
		"E2293D6B2E5A902142E1C4369E9592D61CABF2810911204CFD9C4DE7CB2A225CBEBCE8DD81032D0D",
		"1F348030508E62759E1A7F66000A2204510822A8564A069B9C62897594B46081621180A292022021",
		"A102235D3CD40A30A9A0C11833A07C10FFA01FE01D655C4D849C21910D34D8A0C208834DC1011802",
		"C4149C721AEDD78926C74102884376FC122140226543474F6075B3873080E0548044D03CF75A0A02",
		"08601D0D5DF044913839B173E239868D3142752D7040C43C1611424E455FB0D0E2634958C00E1703",
		"8C30800B2520C0A6152C70115E005EFC78A722F781C1DE9F84166AE8A18826AA440B9779C185A24D",
		"34B0011BF85802E9122A80E00D09202044050B2C3422C9616F38B2D12FD448012A07628A69849EB5",
		"D46229611C34504173E49409DA34608CDA99300699F20063AE1536428C4C5682911D7594E9994524",
		"6C60C01AA474EA67422CA478A58C783C44C889653EF24901D786141B3457A6A4FF89573A7A41910D",
		"028CF0C27A85CD314095D8155041050EE4A8D44B784C531D1DAF51A24D49803CC2DD1F9A1083C709",
		"2E40954E0AAE0E015B2B9E1D790AC33F7D00E0879FFC71D744515E4B01A84C8C30C2030DD050C61C",
		"3C69D5537E0AA7B1DF270F0C25024B3B8EF3C53489A86C0B634E1A610347CCC83A51305C300CD41A",
		"00BA040A1B1F64D08C6A4D91F6C2182914800C7CF8107C0407058000DF06107E310676841890894A",
		"6D14350A296FFF64CAB3303DD49C0263255074116328B081BE318C119730F8D9028887CA7C30D47F",
		"2BA9F6071F776C1BDB68291C5D3112BE6C718705305044DE1E5DF4DB8CE30F7CF0B44BE26542821E",
		"65A1C0D0A04B7050C90705354C53394FB585E5471B1D01851220CF08F33022FEA420527514E9EED5",
		"375DF8B4DA58244273070A9E71195236BFA4EB136D3BBEAD114FDBF84C3B3A2F5ACF36E98F442088",
		"57E10473EC9A7A4D12CDFAB38963254F34A0A0FD6B9378C32F8EE799CF988924F04A936132670104",
		"A4400529E0910B02D5A380ED22460A7C0CFDA8639E8984811C28BBD410D8D1822FB5805113DCA0A2",
		"28F0023E0C221E14B093088D7099F0A00D26E753D40ABE968006E45051796A810224400B1BFC6F86",
		"465000070010428504010021FF0B4D414347436F6E2004031039000000015772697474656E206279",
		"20474946436F6E76657274657220322E342E33206F66204D6F6E6461792C204D61792032352C2031",
		"393938003B"
	};

	private static final String[] TURTLE_06 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE00000022000032000000004FAB004F00004F00004F00004F00004F00004F00004F",
		"00004F000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C3A9FD0",
		"E8B0A14A90488A91CD46A16CBF602EC1868D68409F04A8C19246B78D7843D0A0350AF840411FD017",
		"E47103240D36276D45081D876E1D5E360372577E097D1F9696091F209B1F197B0978245B5A5C2331",
		"2357820A4C0A7B05110E0D1430312771362D090A2478941F0F1010A01F01197E7F117F711571308B",
		"8C482C2971142D02039077770524A0DD807236843629D04F14312D2A29230D1C442C5ED703140D23",
		"032D5DE6FB4E2C0A2011126050D0E519BF8308131EA1206A17961352182A98A8E00109830A8534B0",
		"76A2103838E2006DA120A010850A365E6434C2824BBD90DB0A50EA1540844D62A0F40C6A60ABE400",
		"FF2E1D38D40098C00A41251452A4F846E6911C5B78F2500AA0099889606870BAFA38871C341227BA",
		"D47031828EAC7A8064EE8CEA6A4FB100BC60CEB1B01289BC1327ECF00C89CD56839FB67E72C158D7",
		"89B82B7109A0F05AB8B1E3C790997020C0A182BF77D0A4D920ACD09F830F3213D0B001A506894B09",
		"20EC32D498430B0988C7C5E9484F16DE708F42F7DA0C99C362BF70E2844A20FC8FB2D96020B7AC2D",
		"074F05120EDCF631D627D3D62FD806949B42F00A6329291E318F13602A5BAA2080411041ACBC013C",
		"8428FCECC29044C0F224283069816F11C9385A6CD3875B9AD8148C0899F471CC567FE9F5C50814B0",
		"C0D9115750A2411A5BD033805FE77D20FF4C796920080108D5B915811C28DCC21B23D54CD4C0050D",
		"C0900D20DD18974725ECB5C7C71F8BDDF217690921100636B2C8C556730D0C12D54EB2CCF1025D90",
		"09D9911C2F8123D26C1BDA50D276CAB1D011092A54E008720052E3942C911DC125122ED823CB0502",
		"B4B0669A51308486011FF034219D4F88224E5273F229E8A084166AE8A1883606210962BA24050BCC",
		"C41661632448B0891A75ECC952036980F00F08A234E64F511FEC62CF22D25CD1420B6DB4D1669220",
		"909A46044016A6800B297833CE16B4C93726051D89B34BACBDE8E718078C26C96B58836879C28C7F",
		"7E215C048318FB18415B0C721B8DC709C7132427C05092B58FA5301A05DDCAA4FF6000714410C0AB",
		"E1C4D02CB985A520005A68D571474E6C69028A48E2E0B5A2100A08E6CC3E5EB81949850D28980131",
		"ED91EA8717039434AE101215F0C048D0D4530F54022AE8566AEB81F0F08071349B9D055CE41A2B31",
		"0F4CEA4686B61C768CC3E565921E081080B687019410F2AC386DBA50211E072F2161112D7DEB1709",
		"E5AD4B954D68E43C95D4A0D832EF08100AA2E9101C8CF0405EABB464164F345CE18A316AFC028109",
		"2688D0335B472619C7682828021E09206850C03FB4CE67E51E1990AA8B7AEBE998470329A2E05749",
		"5F2960890656CC77362A4A16A0898726537549D43F0F0296C7F79AC34203096C5040051096040E37",
		"D0DD9880080F104BCC5551F005C2570DF44A11B6A9358CC03270C2ADB56F019E5862CC8D3B75F4F1",
		"098132C2427E15D340474827A4B80D25A0D01D8A1EB1FCE1D709027CED461B2900C0D35973145723",
		"EAAFECC464BE3524ED987C4FCD38E31D4C86036EB082F2C2946281966CC8C56E8E000A9F944201B2",
		"04273E650AC7AEA2E7181BA02205EB6046845C5012713C220C5E305F4696438F141046051BC907D7",
		"1AE082DE256A082C80C41508310031BD3009D3284A004005C01B4663102F238ECC7C780405986617",
		"B0A1C00B4488280E38712541000021FF0B4D414347436F6E2004031039000000015772697474656E",
		"20627920474946436F6E76657274657220322E342E33206F66204D6F6E6461792C204D6179203235",
		"2C2031393938003B"
	};

	private static final String[] TURTLE_07 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE0000002200003200000000EEAB00EE0000EE0000EE0000EE0000EE0000EE0000EE",
		"0000EE000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C3A8D9C",
		"94AB814AB19ED861E35100750B34D59095B299CFA7C1C9DC2025120512C996AD0F3923DBA941D953",
		"1A09010582058501201F8585246B3614660336291D12242009890F747642297E0D0D05808A01011F",
		"A71F6F9881A58624A0A08D149C1D2D292970230A0936318EB071A1050920200F0F228982A98A0D34",
		"28B10D79B4441CA014140323356A366A14027BC06E717072B00D2702DD029BD459650229081502B3",
		"0867EB1423232D0D2A08DE0974C2818402040924381A7165A0C3870E397460C142818A864F580002",
		"1120C183001B1A608478A7810B12270534B0D1660E1A74A15682F2D66D921117652CD87CE7E80F4B",
		"FF6124042D0B00418408441F4A9532508865BE06305A8C686085E71A99A15A15C860CA1884AF5F41",
		"707D3328429B99235E8CB4D32051020730B08583B5875C4C45A4400D2AE000D61F1A2F18904492E2",
		"D1D59FA022ED39910DD64F496B071F61412132111871DCBC5AB353B2E7CFA0437F4E615081E90773",
		"3EAB209142D7163E4FFE1834508C6C67C9C212B076C9A6279F743FDBB5636961C00BD15A5CE81187",
		"8E441C05A12214980213548B136B660D3128C7A5652C2CD8F0D953A1818302069406E05AEA4DA957",
		"7E0708284EE164044203BEC71EF087F9A056429D62CC034511D2CA2B2C3172C608B7D112DE2FA0C8",
		"0187521F800016040924A3082BE85CD5FF0E4F380DF1C8003039D7407A199C624A85A724D31121D2",
		"A153430D0DC23302080B9CA240248F9CD5866EE6FD1748862A1A22C866B04C331005DC55808D2FD7",
		"A804CC308114A25701D0F9180B768E7846191964F8F5488F0D9058DD96AF3C021972458427C0708E",
		"A5F3879A666453A368E15980400A2B4566CD542E1060839362B0C9891C88AC32476586BEC3820A0A",
		"342AE9A494566AA91D145D9A85040D98A6C0032269CAC48F151663D69D031140410D8FCC81AA1108",
		"28304C2A2A82E0CE601C54908202209856030A6B6994420B311466C30828D8400206964C18EA6824",
		"44F086692B49A2A63DFC4502DF197A4550CFAB0E1DE482397DE83193378C75A8FF477F2FBC301C72",
		"AF94572D3A3F99080C4CBD2078C25A0A38498102660844829A08D6FB9F2059E1E5D819FC3134840D",
		"20340082411480DB440DDE8802531C7B25C01E261C59C9639D2FD880804198A4C29A40E0FC018C03",
		"152490D57A482565882871249BCDB15275774E03029591EE94FF0969A1094531F322C23499D1820B",
		"2344F2504FFE09A5D458C41483087B0796E9580D994EDD8E1A27DC3C4C29154290A87A785939D34C",
		"16C3E38D94331B9DB60948AB9D0A1C83C047C3625D42145E5D2740E31C7A5757A840586A7774B6C6",
		"7282721C357BF25903090FBA8CDD94A660980A57A8B067A5C666D0F0887E4EFCAB800654D26003B6",
		"E27067E02161A5A2DBD45EB17893CE0B712F81E829310A10C30B54A453CE95EA1D92D450CD006E41",
		"B8C34807B50D16D0A5975EBAE1DCB63090C342030C0F0DABC02B353CDD073AD0DC55081CA1241081",
		"59F4F2E18D0DA863AA429967B4EA0889105617CCDB7F8884197A3710E6A4E31A58E98F78A8B08E3A",
		"518A32672041FA5662265820C827970A0F6B5C50329534620DEDE0437D1845A98A3188342490C818",
		"F824150A5C601F2950E1A452F0827DD80209FE1881890640800F890A092C80C62E02A081889DE087",
		"4BA8880D786580046CE0011478C1F3908884F19D80112BA3A2A182000021FF0B4D414347436F6E20",
		"04031039000000015772697474656E20627920474946436F6E76657274657220322E342E33206F66",
		"204D6F6E6461792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_08 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE000000220000320000000083AB0083000083000083000083000083000083000083",
		"000083000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C3A5929",
		"9242015234482CA796C821202A14CAA39132A61AB6B4AD4141A74F8D8645F02A1F392ACAEB65CB6E",
		"8D562314277068710505018A8888010909201F01887168140236242E6C23240D7F4B36022368050D",
		"8D8A01191FAC8E098B8DA6340D84B43676A04A296B9587A6891F221F100F1092938B8A2405368536",
		"30B948241F0509062414312F9503BDA7A705CB88099E95850D9814D1EB442C99292E313608081469",
		"0D23821728602DECFFB958283803B0A0C183EC385C01010944810F3610DE211101520305244EDC8B",
		"6363C0466FDE0650187142C0283B12486CD0F0218102750059A0E0D5A681830A09C04982C05384B0",
		"FF04A8AA151050298D1F84BB885622914A112B103CA30A83152082A942F60A525831E243830706D2",
		"84B9726244382B8C80324AF48B529C01846E21E490118EC734291078EBC8C61C2F7CF8ECD9181151",
		"A2160A035CB868818B480A1A245E1970E449EC51C39833237457834263CDFF1AB06AF9C0E265D05B",
		"B21C4071A522504F9F25D64BA1E0110812344664D925AA26DC5A6906744343548D91AF5F6D3C289C",
		"8B6E0ADA244814371727FAA90893187D2B6943B0F08E9D1E515251D04DDF36B3AA350DD092EA5A53",
		"1D3D0A403C22F6BF14F369C1599B6CD507100046924C00E534C05A5686A570C258E064F70A7B5199",
		"005524AF34120159A4A0915933021422FF8E220F4A224924AC64004B0334A070423727BCC0404136",
		"B89082041C74608F70C355074E22A940526253A6784383050645C1D2071A143080003028B59427D5",
		"440654250958D54B771AA521D12D183550830B8419B2E01A6B60B51756CDA0611F6A2CA880891A71",
		"9CA31F2DDDE4535F8DA8E97202095190704475CBD850839679EAC24414CA559180041EE956E8A390",
		"462AE9A494562A0407514CA1E0699626916980205CB866A744D4008923233247EA117A352041042C",
		"3102136A129C41A811355AE0991AF784B38C74A376C0E91F5174121D6E141C05851A97F0828D286A",
		"C4F026120D54700246DD45D3C0061150534565BC760712487450408247645C76FF02080E38B4DC3A",
		"7D8A076E4DF714E82B48599A64925C18AD240919098DE0420595C9790588BFB0D7522C1CB1619248",
		"161C4B0202367080273B24B480499CBCB0C5882A4D55A85D33DF89D2C200237824D14C14C0E7A54D",
		"0EF0F89F08218CD6E32286CC3BEC7D7C558208538B98A8C0032644F5410689BC52201A836E389837",
		"3CF2086202AC3C989D5BFA6934EB9661FC75B553C3801002D54039D2C8D20B7A869982C27578C5C7",
		"EC81204231454320203204C649023FAA222496CF5499F81F8076CB6D6276BD9CC01AC086B9430809",
		"36B026CE64190898C0E182AB15242D3A060B6FCFB5EC58CD533C3D603322262E43121BCF18448102",
		"27C007820DD2DA53C29D4B36A1D2D4BF6B79A391E74EA430426D4EF1268F81B544F68BCFCB5B7125",
		"471EED0C4ABF2C19D0D92506337FF515AF2DDDC69B7DFF6345021B3470811C1B17B5B82FDF8C4B03",
		"C97849C4828CC20FD0190D1E95D51D052E60D9303A68585251DA103EC3E021057E60410BCC113A07",
		"C44929237143775C20005289057C1C7BC348EC94828B55CA1E13AB40034E0398FE51800098B0540A",
		"3C9382169C20360CB842745C5280B2582A38C18A42435AA2810D3C003FC02355B54810A65519F188",
		"4808020021FF0B4D414347436F6E2004031039000000015772697474656E20627920474946436F6E",
		"76657274657220322E342E33206F66204D6F6E6461792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_09 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE00000022000032000000001EAB001E00001E00001E00001E00001E00001E00001E",
		"00001E000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C2A6B0A",
		"C54301516C6CCEAC1669B179290DAFD8061B638B36500414789CB7C37421912890C0B686BE410AF8",
		"FD05057F7401050D850961362E245E3570593614787A828001191F9A01099C827582117C36242815",
		"6F904B5F797C248106011F1009100FB51F091F8505BA0E7B7A5EA94714082E1ABC0A14363131BF0D",
		"86CF8176A0CFCECAC0C2D9688B2E2E31A81D0D1424AE79C1DAE8E9EAEBECED427C0A06207509E046",
		"64E79004600F0DFD356FC44C12E70CDA2F815F2C50B040C2059F06EC1AB8F042100C8A060E1C7C0A",
		"000182898E103E64E8E487C41E73F6D64902F3AC4FA04E1F407CB0D51144484F851AA0F832CA5D1C",
		"FF4904F5908882511A28578614389334A6DDA1041B1A5C68C06062186593268261A90CE8D63C0E5D",
		"F8DCB2A858CA38611E3C40A020CFD8B770E1D888F2A051DC740D12B0C9A5F7EC5D2704F228283038",
		"0299240D30F8B3E1069DA22302BBA20C236E59C3AB5DDFA9D9051112850A0FF41698D28822B03CAE",
		"9E4528E44950A09393C47453B64EA28451A428CE09E0EA4F00999DD8F8E934A700858B3600BE05B8",
		"351A0992967081980EDDB5CEEB8B62BCF562D286A1D67E6475FC0841A6A74E118C7B4161772CD33D",
		"7D766508FF9BD3265C96CAE55106633BEA3D8648B31B2704CA64083976EC81955F7229534318207C",
		"0146232DD591D77316E6658749949994FF193B7369F081060EC43091057A0C64C3452CA5E8E24A7C",
		"7435917B2E28E0C224927540CA49615078124F375E604305B4FD9584171220D059113110D6092CD3",
		"61F04091460A2146955866A9E5965CAA935C3E5D3AD14014529413661369F0F5C06F0C9E29440D7A",
		"E90582206D9E39210624CC99407BC378D10F987799C14A030A013A843F0DCCD9185C8A7017238C57",
		"75F7D8103584265C7FC200C45818DB0CD46225CE34DAD5A88C28400202751EE946680A80701A45AB",
		"0C6A8743D524784D6CA43804C38DA92AE18F037A8DE9D017E4B0325480BFCD335D497CEC4459ACB5",
		"15636369CEC4B74B02195852C7700674D21D34863A2691A3FB5542522C208820FF42BAF86DC46129",
		"958DA5080586E45141058914A28B146A49A1CB2EAF5136A94FA6ED010A749CD041C8B5CC7A888D7B",
		"030168897DB4A45B312E04F2664DAF90AC34E8C17EA46B0B79E51118A08261604AB038D790736E2E",
		"204C618B5EFF52B394CAEEF054896F9DE4C2C97CAEF9C69BAD0DD480733BE63472516A82C4349D2D",
		"B66CF24782950D9C738BAC24B88B6FD9069213D5405DE99E0D5D5C23D19E4A051C8D1E09449008CA",
		"1FAA43413F1FF863C02AD8B5F20C21C561C8A11E3BC5ED65ABB16C90486CD50C5ACDDA4B0DEAB13B",
		"63EA698043DF287ECD7B020FDADD1737FA64C25CA616ED420D5D51C8122B0E9844214A6151D0F958",
		"9408CE94A0128A2131D1B81404C6F176A756B0E410133EF780AB8D1E6D242908F049447272CED32D",
		"06E1066A01AA623EE8969BD867AF7D2A41000021FF0B4D414347436F6E2004031039000000015772",
		"697474656E20627920474946436F6E76657274657220322E342E33206F66204D6F6E6461792C204D",
		"61792032352C2031393938003B"
	};

	private static final String[] TURTLE_10 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE000000220000320000000067AB0067000067000067000067000067000067000067",
		"000067000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C165929",
		"12680AD290524C28D6C965DA04279BD8763ABD28A446794011A7102D92E23AEA1A2F94C6A3004A34",
		"1A637F80270D72058787018A098A018C01878284025B764D0A6102820D88058D011FA11F091F20A3",
		"908A11056A03802F965D3680829E8D09197D100FBA22A190052491926EB04D508002797F659C7F9D",
		"AB9B276C0D0362142E14C54C29230A731414D5292C421408E3622E29150D463105A4A509110A1227",
		"1CDAF9FAFBC52995FC0091D490636014B006FF88CC21412241822B018FA4F8326D93454E6904D900",
		"1766E346802938CD51E082841893B238C16B9401020413104040F8A008510314164F248CF8254F4A",
		"FF4F89404D7119D354869ABF36D5A0912D229189842C023BE400511A61C032DAA0D1A08D0DA71D04",
		"2800014A4301182E6C0C88FA87828011025A5DECA8A9C108700450D8D80956090B0AFE5A90EB4BB8",
		"B0531BDE1EA43931D8B0BE1309C892F2F4150902310F4E90A0D0D8F110049A1806ABE076A33F2335",
		"1E4C5EE598438A1827FF5453134D8C00011429DC4DA6BBB336151C9B0A19932C0C2D540E1F052BE0",
		"00908DC5627CC3022EA14F0260DC066DAA155494A8EBB502646426C6A982110DD2CACA9387EBD55A",
		"19BC33EA14C040F34D940A7FF9F3333C28112EC1E4926449A1C08C1A2F3040580AC9D0120C4BA0D0",
		"04C25140FD42D37DB39447581B0380F1FF4730CE38F24823CF58D44A35C2F5D5535751B9F0901F0D",
		"BCF84B43D7E574825B7CFD36C272246850811831BCA0911A5C29C3DF915D75A50936D8E4A8CD8ED6",
		"4540DA0817AC9552546564691B052AF8C341570C25F00008C054A0401B11B9A0809343B401D8082A",
		"20C1820A2A2850430D74E2E3191263ECE9E79F80062A284F032134A8251CC831160852B279E85334",
		"8E324A029525B1511829FE594364917D004FA54620209A688E829568031248E0D02A0922A1428F1A",
		"8C428205818E30862C65C036C68D5F0DD600A307F939815779E0060E7F3190E042266A7538C08EE8",
		"3510A7632A88D1A118170D43061BB8DD38820B2D08205D311C7499D0440D2629FF88B2CE34905403",
		"34C862037A62D410500D149CF0400335D8A085BC5F68068D2753C403894A07B385020AA53281C203",
		"369812D91CCEBD95517841D53222899BE0F4913E1CCC114104AB16902F5B9C30F26000B888E0B2CB",
		"1F64009E30F2BA10434473B820412129ADC7950355811280020F8C39454DDC6D32C09D0D7731C75D",
		"642CD38003159040E222C9FD829478CBD8808249844910F57657CB1C93083299E24824AB949107D8",
		"846D64A44ADD89B08B80108870D4BBB4352040AB7D4D144E54F445488529310BDD08881A811A110B",
		"EBCDF6E0217BD394400896D394C841B2A9E56FDC71D1828801FF1D3E852FAB0873C26DD68CCB0F54",
		"483A83C82790F0A8BECA26E8B95918E475F5ADE87BB2FF715D467FC4DBCAE7FAD960C1A50E2EB7D2",
		"54D0F087A5DC4D3B31E70828D43042040E7CE161B6E0B742481A64E4E182531CB4D0A559A3744581",
		"90EEFB2D6F1BCB78459B1877B54001BEF981D44DAC1AA05A0360701B35DCE6467938506DC8B02312",
		"8C404B576803AD40A6004F7C035F14D08B3FECA206E280033645789BD548170A0CA02920F3DA4C57",
		"F662291220A0017A2242064755B4344C6B502149410B04F3A82E40CE02E7E8A1108748C441050100",
		"21FF0B4D414347436F6E2004031039000000015772697474656E20627920474946436F6E76657274",
		"657220322E342E33206F66204D6F6E6461792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_11 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE0000002200003200000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF",
		"0000FF000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C0A498A",
		"2824016AB0922EDBE9617B50AECE30F3602B9B4FB601C9D5300B2894512B4E4989EFC3547940D116",
		"1A0D0505018401098385090124800D270D03231678481C36242429128C8F900D140D15158A018201",
		"1FA81F2010201F86A57F348D0D022F93944D146FB32483828AA8220F10ACAABEB39F9F6FB861BA02",
		"03030280368098817F828D688014686C17287D36762C29B7CC2C2C360A20051A05159F1428E3E536",
		"08427A242A231CCCFF00030A1C48504C0D12201268F85080D1392315146C30902053C1242C469441",
		"432D10B26CD338326A63030C400E9814246087C9CC8059BE107D1001C1040476AE4C111A09880689",
		"FF921789D09B0648905142AD1E801886AA50A1020E1AF5116023E8107536FA9C40E1F19A51907F1A",
		"24A8880C8D0017032EB25021CD01421203D2740405884FB7470EB89E8063A703234C8434248890E0",
		"525004F898A418E0428549AB902353B21191535F8C7AB25E965C64D387B1093E07A88A44019408AE",
		"12C0E07C04410312180A28280087AA191B73C6B970F1D70A6B22EA5EEA3A417C961979DDCCA090F0",
		"BB831E691F018F0541289001427F5E722D53F0848206E1CAF569D411518104194EA10AE12ABD51F0",
		"DC00D5D84C8982840D093FF722B16D1B2007500D92DE679F29320829011810563223BC70112359A4",
		"01D321D885B64A4DC44020427AB02CD2FF1079E358A5471C6D00720822867C90C1582BA658482F7F",
		"9C20806D144446CD4B8D7C3588684E65534004D60032541B8F5DA48771D554345B20069270883670",
		"9014C90829205064413FA5011D32B2BCD491366750D0C208FB25F09A040F08A0024014AC90C2421F",
		"445003343050D58646E49935421C45DA50C377E35450C39A057100C20051B800475648B070422627",
		"3437107D92566AE9A598664A84045A50AAC49592A630023B20B4B4440A1428F0286995B200454220",
		"B4621112246CE264459E4AC68258D4C57A9E6F48D8004204D4FDA180A5DF91B0016A3FDAF0428D29",
		"5C61D26C11984602A614C8C2532D1D9541013436BCB1E708E4822A9956653C33FFCD0B71851B57BB",
		"6DE466AE551CBCA00579C848F50871C4816286A46DC0D149BEA67CF86B1BFBD2100E6B2CE812E596",
		"11105220826395C2133565E44A89027C1171DBBD98E858484E288244CD23ACFEC329059844451235",
		"983CFA944C22D44C8C2BA5C0926F5CFF7040801EAD9C97C9088E0C7CDE794E2DA5542BD81D35D246",
		"6868DC0402BC6D70C858BC9581704F003690F329A0A548B2513468DDC60BF38A81D208A268F42535",
		"0D211D4006197CC04A86225077E2825A8042CF451C2850F696453935374D19B2D2CA8AE6B9EC082D",
		"3506F52D5D8DC0024C52B1BE629E2F3CBD2C3533658CD0C951A57088CA2A4A355514DF9F7067953A",
		"708C474221D76510D86B2B8A64F05E2FDAD4C5286429CCC80BEF5FCD9CCD7EB34CF5EDE7FF6C4470",
		"90D98CF5559EAFC545810B2B48166EBE6486EC1548F9D2424D0B0354D007F0CE464339DCE197E82D",
		"C60D901B8D0065933055CAE8FC148103CEC6004323D2D09A84FE639CAC7C021309D8C04240B001F3",
		"A58D092C40540318F2197940034C514299E788C081019CE7760F480123FC0110152044030B898001",
		"D8C0001750A50EF17BD9595250816B1D211D1DE00009B1343B4344A02E710155C36018B94CE1E627",
		"24484CA346808016304F53466081BD8A08C52A5AF18AAC09020021FF0B4D414347436F6E20040310",
		"39000000015772697474656E20627920474946436F6E76657274657220322E342E33206F66204D6F",
		"6E6461792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_12 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE0000002200003200000000F80000F80000F80000F80000F80000F80000F80000F8",
		"0000F8000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C95CAE488A",
		"0729D148291B52D027506834BFE08EABD150906DB61785E4B20904B6011A2D4055C2F8E329DEA090",
		"1B05010120090109815C88055D3464232814365679452C340F050A20200A2D670D27277F0E0E8B8B",
		"011FA91F2210222086A601245D673691944D1402147E366482850519A89B9BB08B0D0911247F0D72",
		"032E031CB85F68CDD705CC80D77D23A068DE6B24821A090A2427D3D442082D2B29A81A0605DE0CBA",
		"92422C432936242A27FAD6091C48B0A01116916C190C238184B14D8B485050328244381B0BBF446A",
		"40020E36640D1C502183A24149340117A614E22B942F45813280106102822641A602198840E6C480",
		"FF6F16D671E0A0C00688142EC6B4F145C60F20415C04CD04F1404B21A83867F5FC1969524621FC1A",
		"3025436256CE525D12A80529F6A733010D5CA02080C2168B395E07B2608626068C3F4E79F5DC2360",
		"040C155FC13CE0358682BA240857269E4C9909BF142C241BE1A7604DE7CA4838407910250108894A",
		"6C2848302FC544D046CC98330062510539D6FAEC5263A14F0B01B091A898C3C7592D1BA27CD96276",
		"A26EF0212D5ECCF9436BD1165306B8680325B6850DCD19230DB8372227A20F09567D18165596833F",
		"364EBE165A86040D1A7BD00CF0465616D442ABA1A7567A6BB125167C1851A2020529281001810994",
		"851C1903FC54415938A1F20104109850FFD326EB5D35CC76DF7C67D0500D5430E181801C524021E9",
		"AD6788080F88904A22DC8C27167803A5B04B2F1C29921D7AFF9942865A67884282352DA480405E05",
		"DD654B28A2909149845AB948CB277FC035C218274CE11006FB2DA45F0C3A96F447736D5DE3130244",
		"C4C0113A2C3420C14208D8D00527CC8C60430C2DEDD7932D6E908040033C52C60267B569B04C1F72",
		"80C782002D3CF9DCA59866AAE9A60225CAE912778DB1A4A78A56B086370922818038B52D066570AA",
		"69028169A91ED1190811D4964DA67B29904A7A819888040B0EF00985A6615A551B156AA4E01A3FC3",
		"45628144D1BC1A5C59129CC38C052F54995F1C158E40E92EA44EB6D41CC9BD80FF9B2F148CC7145E",
		"9A6ED480476371531772EFF92280B5934D004E28565A794A000D44F5227CA00810146C4BEA48DD2C",
		"8608F81F54DAD43BDF64FC08E61620886085DE8B89D082CE1F7E56969035187251C82AAC40F0C075",
		"2F06A28D003A92F042B94CDCE5A73E5292D50F54E7B1F24055AF60758C584C4572F33AFC8C70CEA0",
		"B91D69DE1619A815B116C31CC2053748F3BB0407A212924D8409B178DF851D67C0724D1CDA98E1D6",
		"4DAD7B3118235420C10687405116C0406673CA8B2C3FC02104AAE0244C44629D3C771E4369D34242",
		"4E51178B86C5B01723C11C03E68C4219DDB50B75C90C0C4C2A8243F08A8B65D1D21C53E37D252505",
		"0067334EC71F6881D3553602738306052D3C66E67891E7C4B179996B45C6E7E03C6EAAB0292C2C90",
		"2EE3895292296551B756F5D7F432C2F61572478102298C5090BC7C399282EC9977590BD203B450D1",
		"141A0420CF92BE0BC48F8EB00FB60D8572F8F11E8514D0C7185E44080528E042380B83940855A503",
		"ED613F7BE0487EE6E61A845486191130C0336806987A81C3059EA81F686C509102C46F031188812E",
		"6230823D08865243B081DDAA902917A400041BD0C0C12A50035E08CB08927012A7C0B70803868F77",
		"89FAD30BBC7629D728E053508CA214A7688420000021FF0B4D414347436F6E200403103900000001",
		"5772697474656E20627920474946436F6E76657274657220322E342E33206F66204D6F6E6461792C",
		"204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_13 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE000000220000320000000017000017000017000017000017000017000017000017",
		"000017000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C94C525214",
		"0A4BC9692812894FE1D9EC7A3B0D1296A420352C360AA9751A0CD2034163D438A16CD3AFDEC878D9",
		"FE0D812424011F5805010105888B05618F36288114167B4979422C36278F6109059F881F01581F1F",
		"2010A9A6A4881901660D69026F14961D2E643665810D2DB136038136BCA18AAE1F222086898B8187",
		"B1C1B1BD2336047F6D0F5CB645149B36B0BC9C0DCD0D14B3BC7F78DBB61CC19F1F1B0E142D140C35",
		"69483616F4EBFDFEFFFE58BC7061250AC03D2E148050B8AC0126222C528C5033204AAD837AEEFCA1",
		"7042DCB8668B5E390AC4498E8D141891748B72E2DB2745090A8930014121B344CFD09513A6CE120B",
		"FF1A0AAC245050A69B0D6A27283470D028512264201E281855C8941672B1EC001BE0420087941D34",
		"09B028AE000947681D3005158AD7AF5FC3D48839F560D043809AD20988E1B6CE309E4753A4B001B6",
		"0B8B06291CA550306044E1C74552DC85FC6F443010A448BCF84AF9C8371020486C28B74482838524",
		"A87536721844CC04AECD0AF0D3AD819BAD14468C453179B5100482BCF9FD256098B7BF9C5A5EF42D",
		"84423012C3468654C4D629B87234CAF57CAC60A3D2402111E1B4EA2ABC4861C1DEBC4009D628A006",
		"37036438057A19EC51A1100DE275FB4FEF2655985103607FB9E5C0598964904102A9404093325A38",
		"F50A38B9FDB25C171C10500608116031FFD408488DE35623A03C65CA7CA83C10533309ECE7D61B03",
		"B4404F4F14ACB7CD4ADF8D3388530618405D23E31C32122F464541D0080F14D2E209FF68624E1D0D",
		"4802DE156280D7575F3D5920C0810E719692581501138894C6A5A154473608409002CC1941020511",
		"9050518D7EC4D2820D320ED18D602DB42904074F4CA50116719613C57F1DB0E727158B36EAE8A39D",
		"1509A91E9A94414236884E4A441544918099225EFA59811AE35C78C4612AC63942A8CC25A4C00349",
		"82409A127295A3C2A29C2A93C522D0B130C513F3A4334F035D29DAA61AB065A145040214A7D71FCE",
		"A6E35FA38791A1C0487F34D6D25F63A561C6093454D4A80B1A6D74E5B94AD160FF5C0C951C1B5771",
		"7D2922EF8AE01458C176940D46C75F2361E19A844E398202382754305B672A00E39159558D874579",
		"E245E04039492965C30BED8295703472F0B2C87CA7A4F2C0284D3123CD24BF646A4B37CD3E474848",
		"19A002C103A031F3DA8F220AD3C61F29D5062521F292F214168540501F292F2932313AD3DAA2C944",
		"6908EBF1226A91389F0833A5521F550564C06B20729863AA179552E922277FD1504105AF89F2C1CC",
		"222F5388225686230EBE5D84015A7E24B444F074002B237806C940608A23625C67610B00B08AE844",
		"977698820B1254400D303906C24C89E3258319C41FC1320C35BDE466140B16A9FC67071CB4CE8102",
		"F7A6E1F147C514A2CC8C53098868964E6608404F37272C021B0636B0AA8739490DC309D024362F86",
		"59437A94540B230C42F450299C603CA59877CC8B196695D12F9030A2F9470AB04C95801A15606494",
		"1B74B8858B59E00BB3132C14842A85EAEB70E451378A43812466B1BC34C1807F60C9CB0062E0B340",
		"44230A3CA1800B80B33DE62820027F80C16CD0018B8A4444132D5803B52A702043D8A60530984837",
		"A6709716A8000A8B3A89025250954F54A0013558211252E007046284034721031666E80B1768AA0B",
		"269847FE8EC8C4263A913941000021FF0B4D414347436F6E2004031039000000015772697474656E",
		"20627920474946436F6E76657274657220322E342E33206F66204D6F6E6461792C204D6179203235",
		"2C2031393938003B"
	};

	private static final String[] TURTLE_14 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE00000022000032000000007EAB007E00007E00007E00007E00007E00007E00007E",
		"00007E000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C1E4923",
		"50221150A59C58A782A4F80412D204E965B39D4E8D41994241931A27949A92AD27298206095D68E8",
		"C10905010105818283110D7D707E0D34652C764C2C65958A8387545F091F22109F10205E8483248B",
		"7E27360D142D140846169149366F62A6247B688A8D7E82058719019D2083016FBFBF7E037E66AA6C",
		"2460242B0DB2924714650214236FBA67AACC362314CC0229290D1CD6EB422C34BF1A1B0E15230D35",
		"D857ECFAFBFCFD1D2F6F1428B051CD1F960AB8486858A524058501161A8CB361B04E053DB85229E3",
		"D52B909846CE50B0B191AF62116C6C1A21FBD5C913082984A85041668C63834A05ADB1D85969192F",
		"FF1A15224CF9C5094408100FBC7C7890D45081086F18316BD062C400027EC04838A10E490B170D52",
		"A8C0A580838276360650E8A6B280A95DC756728CEAC224166C36C80C8801EEE6897A36BA76608560",
		"84DD26235CD0FA35C0C588B387F9EDA413B922A5065D3041AEDCAF82DB37DB381F492110E9879904",
		"DBB1112060C08017372BA5163D84C55A400614246820E0C5B65465E0944935A082AABFA972567681",
		"A212365EC50A4DC1E448CF54892F2C54A6F4AD178900C48A7D188F29107554DE789729C98F435A72",
		"E2A01112040202D3301F5E869F8F296ADA3490F47399757FFCE2C07C0564D009281088E005271F20",
		"838B4DEFB9F0900D14D83690724390B0FF5623F53032822ACBD0C41F79E389F000311924D3802152",
		"DD340237D06CF08502AD24C10109368010C12F294850810B0294E1932EDE1563008212CEB70A1CA9",
		"A44299052F5490A141B661F8904D8A40F359236A35924A0C86D1460477488609E6467EB091820D2D",
		"A8708299A32960003978C060559148A044E71029B4F0400A1F68A041011BE0C286607F36EAE8A390",
		"46DA01256B702829129408D4801430307A699F5C2804825B147DCAC4380F44B0455890927406654B",
		"A4E0580C14A8E028390A3CA01B081E2A37926F553D4483A591A5C08514E315328643C1156983B3CD",
		"D8D01C7B9CB140822853241BC1882790A098967968C44B19DA99F906171B2CD2FF0C6CFF39E2D62E",
		"36A9416C457E19E9532FC97C514A547ED443C34D0294CB59B47DB154CCC11282F4E44D28C4522C1A",
		"489A32C5C4A765108C2F88F4A18A33B1CDE60F076A04D74B040783E152794CE2AB0C5E1EB717DC96",
		"F852D18909F7513C8847C538808A6B0314396F1D94A8D5861E988817C6171F80A2D44C4EB6E92133",
		"D4AE43F025F2FD6234042698104AB2E7714443C88D9580400AB2E0F5F399667CA8481F0E583DC878",
		"9F3C60DFCDE379D107541CA1D4826B040427018F0028C1327022F785D021468BF285DCC4B8580833",
		"DBA4D1CA323628108006B99DB099111E2AF092401E8A7C535B4469823479087629E693DAB4901078",
		"BA92A0040B98BD34BE710A80BD874A238690F2B6797D78B99BC2C0A9E01D2E4D5CB0D6061BF01A68",
		"BF283C92924AA65067B5CE6E39056523AC1001AB3F948639802E5F16007A206BA342CEBF18D2F95C",
		"5ADD21A91248C19931420B0D344A29926D74FB2F23420B870048208140392A4B69514B91E095065D",
		"948A12247048A4D6D213D7386727056141C0061035DAD0CE73BB8941032CE0821AB4AC081D140DC8",
		"5E248C422D422DA53215111C2310CF0540206FF2940C8BC0010E98A05675D9A1108748C475040100",
		"21FF0B4D414347436F6E2004031039000000015772697474656E20627920474946436F6E76657274",
		"657220322E342E33206F66204D6F6E6461792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_15 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE000000220000320000000057AB0057000057000057000057000057000057000057",
		"000057000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C761003",
		"852292A892482CA7D6C96A3C408A042861A30C29A7D3C0661BB847A486BC6113C0CCDBFC91CE2E9F",
		"E60D051F018405050186880106110D28747336277D297A4B246D0D6B050988871901091FA21F2226",
		"1010A482198657098072360D23176C241F1F110A361C4D2C36237D7F930D02812486098901190919",
		"B8CA050D249E056B14732E96431C23145859492C14B26E6F737F91D88036422927242A03DBF4432C",
		"34CA1A1A11D20D141408EA091C4870DB2F1B0AFE852BC850C9BF59C11A6E6910669486310528C0B0",
		"11634D99698738C14A27ABC18917302A492C52B2E4A1008360828000620CA298CA943133048844BA",
		"FF7F78548EA3A0528BA47472A23968500866A90F63C44010B1B35F1C58C5E690487085CB37651FB6",
		"C611D0879D5664C88E29103B47801B3A274654A8D06BA52F727D6CC0705B876C51BBF43810BDA260",
		"5C5DC088136BF945E2412765241433A450832B096F78241B61F16BD2C2222C52B0B1802D2F9BCF8A",
		"539090A2AA0026369106F061A3260E5237272CA06E386EDA6A05D24ACE69D5C080814E8860C9163E",
		"9BC26E8268FCD9E0C9A90A295C9E3A19DA7E05901B0AD79E77E09CA7056D5886408980F065D4079A",
		"22407C78A6286D2C4D7C64B7A8804252030C1234701813A2B94156209E34754829A8B0F71808227C",
		"B01D716D8C30821B2DF826930235E841FF8E3F80D0404305C81062E228A19042C82BC9F4C3C72478",
		"FCF217138229D000576BBD168734EBCCB16221132E15888BFE3840031B9959C20126203890800A2E",
		"B41087021C88E616883D3640C31C236438477F1464D8C7372A8C30636AB0A923005274C8964292BF",
		"C4D08E6648D830461C23CCE20259E2D179040B033C90C218832420C5097DFAA9E8A28C364AE79BCE",
		"396A090C983446020C0A486A14059C0C624DA29A0E612309354570C500A086CA8D6A2EC46061AA89",
		"A5FA4F192ECC8AC299927D7345239112F1CB006E7954120D0290D5AB6236AC36D32822BD809724C0",
		"BA85862C5B96844D0D36E0CA100BCA8E82510415907585305AB988020A145420FFC709C5EA661789",
		"B88060EA2B691023473F9C8424CD2371C8B28E00D866BB120B354CA3491DB2F9989D89DB05606A70",
		"C3ACC36E192BB940405E8008A21322A04C4848C63CAEDB8744BF64D9800B0580606206EBA9AC4828",
		"0912E2C055FE4142B11114BCA02D1207215540048A94B2DE179DE01213CC2FB9C64E1B36B4E0C29B",
		"A751B056924C98F7102041B78C6202EBA172B42758C9418E856E11F08715BC6881C93F7F5C61E28F",
		"A140058109A744D8A9A70D988A4E2C400502C2712074B5986908E65BCD332288F0C0030ACCC79E84",
		"3C55D14FC26DB400C7562034C042A6797C785520726CA59D893561E7603543CA21DBB350267CEC16",
		"07F967EF314923B7D1224D40DEA8F4DEC5C09A041A22CE120B9B1927B84C35C94483950DBE23B10B",
		"A986A2E59324264BE336D8D6DB37393A1FD5339702F25591820415B440010A47FE93E71F2EA5E51A",
		"6444AE6992002EA89082E6DB10F08D045250C4C1801D104D19C0338CA344021027E85B028992AD37",
		"D1691CE4125BC2E6808D01FC61000111020BBA34809D01861CC5320705E464925800E039AF5B14B7",
		"22D0867FC0C002233892AA9C47915138A00217A8410DD6304323B4001E1A8049013630B5C8F4D057",
		"9653CD6AA46139231EF189508CA2141B12040021FF0B4D414347436F6E2004031039000000015772",
		"697474656E20627920474946436F6E76657274657220322E342E33206F66204D6F6E6461792C204D",
		"61792032352C2031393938003B"
	};

	private static final String[] TURTLE_16 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE00000022000032000000005F00005F00005F00005F00005F00005F00005F00005F",
		"00005F000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C954DA2824",
		"D0E75320A59AD8AC7648190D28029B6D406A9C1AE651A326BE6EB59C8622213590120DDB19DD2814",
		"027F80807F7E807C24057C360268038D14081C2A3657360F36245B29627B7D7D090120105209091F",
		"10261010530553207E05097E666714278C6889242C6F5C360D6568140D150E85B00182C97E0EB989",
		"68BFC0362E15BBBD422C140A2020B22E2E68241C1DD9625F7960C0C28E7CD094D7F058290D091A0B",
		"54090A0A0DBCF1FEFF009BB07013B0601309245C9080E1C2463F8358503C780022192812025EE869",
		"D042CC2F147E48040B07AD013B3427068421581001051462C23C6B052AC00711A340B5CA902C804E",
		"FF58ED0EFD1A718100300D741464D2828EC2805F8448182094CCD507293629EAFC84A7111F04FF52",
		"BCA4010C8C9E32CFA2F5A91212ADC8670D5008387782020B1636462018F1F08D8A71D7CA25A0E1A8",
		"0B8516239E405CCCB8B1E3379B6C287840C2D1E380D4B6758B40C20BCBCBF1360D6811C6A318BBA0",
		"87DC8D46A96F8A13294F0F3387B25603B22712BF80813A200B17FABA5D9515C1C28B0131F2EC2960",
		"A0A72112113CA180CDA7EE098F9F89B07831A2C9BC3E22F308A010CD4FB73FA078FA294528D0A076",
		"27609E6E30A2028133A5129050A1654C50F7368960825601E0A48A0854F4D487029C4153D8786A88",
		"44870D8069D11105189E508002CAD8D4FFD329AA40A0C00719E0F413490DD8E288180FF5059969BF",
		"DCD180033539E75355A32C038C2C7029428141A6417346279FDC68882E7774E5CE198B58314F7658",
		"A840C203DB28550038306238CC907CA0604650B79811CD8FBEBCE0621614A490026572548844647A",
		"A444410CE3D169C1134F41991A139D9912813115A4D8C09E6F509282022394A1409A6E12EAE8A390",
		"462AE9A4589CD61AA55958A01422B2F08329136295614A0065D8F0E9124A61B229627C9DCA844B6A",
		"7454439AAE1E81D70B1624B6D12F61BCF0420A673A062C12789137C653278991C71A5D686401038C",
		"CD03DD06AFE45109274FCCA6161FB82987120D346017EC162CC8B1CD031FC842FF42052A9191877F",
		"880C82485025D525A647BD2161C1B8431C3AA24FDC74AA51348894616421CB445000599EF031C017",
		"E21AA1A59E432094A4021894FA9F7308C7120B8FB014D00C4A503082A10B2A8815C609246C20C109",
		"0A08B48F699DF8916E2B377DB00C28A50CE2611542FAD802798ED481C009FC1E31C21E675C594045",
		"36894291CE577143A2CE7FD0DB8E182350E0422C362915B316B690A096C2056490C14DA374538A2A",
		"52E86C0AD6057056EA30798C3642A262F712837FED20C39329A688924A8E5789F05E2C60B6C1410B",
		"30A0008F68628AB4CCE5695F35CAD4093C00C132212B82AD6F8B180B4C7B7D7478D5D574F4744C19",
		"4FF121804178A1D8D44E3581801D004F6B83FC315CC3C0F482008D86B65149CE1C13B2015365EDC9",
		"2C18BE2B400A0898FD0F0B93EC2A8CA0FF91003A3073BC8592A0A64D9F420B81FD52C3862984775D",
		"4AF08DAF352D7C449F473F6A5A80612F0AFCA2C0C7FAA9800BBAC086F1D40543ECD8C89048E02541",
		"B9EB049F1948606A3082FF290C01C0388258F41086D841833C2F605A5D92D60B0E10605D0DA098AD",
		"E0543EF2C4C0056C20E1A4583087FF3184023578980C25B53E9C25C06E14E04FAD88E0023569860A",
		"2468C1A0865804040CAD7D4C8CA214A748457804010021FF0B4D414347436F6E2004031039000000",
		"015772697474656E20627920474946436F6E76657274657220322E342E33206F66204D6F6E646179",
		"2C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_17 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE000000220000320000000037000037000037000037000037000037000037000037",
		"000037000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C934DA6C8F",
		"87A2D15035AFD8EC3165A3909E545BE374A2342AE65703C64EA55857D67B494929408940A1402209",
		"FE1403640D7B05797A0588887B018B546334661436232317368606200F0D5A080762248F248D891F",
		"22100F1F79212626101020011F0FB27B54058F830D085A4A5C024F362784B96724898B7A0106850D",
		"0A09A28F2E634F2C1427292E0329BE426234240A0A7C2414C1282836C1021481ED02C40D03ED3130",
		"022F0223343670DEFF595C2858A1A08FB80EFE002A5CC8B0A1C3862A2AD4DAB0215AA511091F6A14",
		"F3A7CB305C8FA8481B4346DE80311D8509E0C0A1C593110D364810F34BDC07102022442870C2C649",
		"FF5C05F014F8F0215629A21F3264F8D02800210511A4F96C6046CC231ADDFE4D12F0B301893CC90A",
		"05587A1329D2525E115131D3C0868B0A24B2361466559E1962B98CF1094992CA4F982D3232940017",
		"6A54121CBC41F9A0C18001AF234E689C4CB9B2E5CB987DBD50508B6801046D333F94A06084A81693",
		"84D5912B3ACE08AE7F0674ED6BBA018A47774F0C785247890D2E4C5848EDE2868BCF11C2702DD3E3",
		"356A8406348A3DE24D7212854A1508D43094A0A0822436C6E149006295288ECA9717C850EA5053B1",
		"B7F83E1A70DDEBD7684DE434A8313D39AE558D9C12CB2A45C1521E7B48DD9240027CF56443432CD4",
		"F5485E013682202C0F0C9840866825B0FFD724FD090610186D51D19E030D94D2948563ADE28C3354",
		"09D2451F29B4D00243115A259F74619552C857C684E10E15021091C2240BB1001331C1D8C5177A61",
		"8C51A27CBF514615907868E31512C5D1E5257D66D450C300564C4680390D24F68F0443CDA281216E",
		"90D0DA42ACCD69E79D78E6A9E79E7CE6790D6F29A8A0669F4CBCF085217A3C48E812D710E380061A",
		"444302378B26C101152718648E6C148858A9110358E0424A35A496DA6F0CD4A92717B23D31803428",
		"1043CC6D637C61033F743171CE080CA5D0513BBBADA30B4964CC560D5DAA0E01C549C1CDE1C430ED",
		"84B187A4D04D685B5B30E1E6171523B8E0C639C39010E908C90A91023111E0F4FF412EF884B21550",
		"8D80954C23A38C14929795A030C903B3805081A742D470C726371562410C36A80114230C17E27022",
		"ECC5C717720D8C40810B7D040082A2490877075309447A1E5F6D26920051CBB8170058CC1122CD09",
		"FA44764225F6C5C504072990609E04E77135E18201A282C7CA388920025343B1E794875261FA8410",
		"A97A73A93A1E89D4E6CA21406042790B82602053B3206D3591C296E150175212B2D37A1A43A00081",
		"5E1B38162CF23A2C5F2F0D1927CFC2012A4D140422E0D488512B37E0A1C40EF41443270C5980765B",
		"192FC3C88A613702C2D18B7888E9DE02A4800009D92874691781849248010EB0AC877B87EC816290",
		"6D3DE1E01371794BBA278921D15083030F33CC5C2E05F541853A6C717CE46EBD8A517CD52467CC87",
		"31244416524F549439598E62D8650C31120E9F3609B486710202002B746438061DEBEA75160782A9",
		"3460AEC3BD4F1C3F0405093819AE400523B4D0891CC619868306211B84AD83023110D524CA059023",
		"7DC53C2918C177F2F32D1018200505A801092E308F34510601E4DA405C0695854BA5882809D88002",
		"CE55199CE1AD81330BD9063ED0161790F053CA7207B92688C31EFAF087400C626B82000021FF0B4D",
		"414347436F6E2004031039000000015772697474656E20627920474946436F6E7665727465722032",
		"2E342E33206F66204D6F6E6461792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_18 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE000000220000320000000084000084000084000084000084000084000084000084",
		"000084000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C3A9FD0",
		"26A5D1B0D51A2E129662B3726DAE468542B295ADDD6FF7D8B5354009852D5AEC56EDA44222B02F04",
		"027E097E7E011F01795489366455530D1431734D360A0A0920200A18728B890D01207F1F22260F70",
		"01A426100F011910107F050D8288540E5D2E15557446169D55547F800E0982B10583800D052496B2",
		"9E5D65BC4E766A6D8B5C14535F8DD753312E178F0DD34C140FBAE5EAEBECEDEEEFEC6E36840508BB",
		"F0463092EDD88B6D8E54486899224F0B953467DAB808376F0F887D49CC34E3832C41040BFADC2C13",
		"F627C10743803E82F818F223B25909A8A879F4AE93C6659F0879049441E4483E10E02093B5935B9B",
		"FF0EB87CC5D368D00D174F3C3DA100F6F2A03F7C426C3C78A0A0009C85E4A2D4F8A0218086506BA0",
		"8A1D4BB6ACD9B2D760643DBBCE4E830816156C81884FAA1B778BCE1844A1911136958D1A1DA5B070",
		"6105026FF83CA8D1F60B37120904CA221181F233810E042A4A7406AB404B74235EEB9251A01D66B3",
		"00F5D1C3A7181F40019629A85CD469E825352A59BA84496E536470F68880A0E0D4AB9C1E0B7C80C5",
		"8CE7B3717778D57073E58EBF80B1328022FEC1E3ABA9AC02E48CD500F2338DB8484840E0E2F69336",
		"890A2888059B3EC70C7B6037B0B4F7AFFB7282C107130A15E881CC6B24FC21D966808545D635E53D",
		"F2DC41D04974903631701103032E7021FF9615FB551506565048F5C1061A24B041256BC123D50309",
		"3800428B509C83801C34B2A5E38E3CF6E8E38F400629E4904416894417E001F29F911DC0301526A1",
		"08E4E18353F2620306A69842820B30B437DD3591B4A10F0C648EC6C49518DCC50E0516B467C74B6E",
		"2CB51460B68DB64807DAD8018201F60CE58B1B2E14808216362C250F538069936827D9486500080F",
		"E418853FFEC144C59C032E336122286CD3571690CDD8966097F6710C479A0A838C419EA804E81405",
		"30A68E1D6E3A128001F95564A06AC91814D985E338F85E011B60C24719D9C410503234D1979F28AA",
		"C62688208910B4241B53ADF2814EB218159F4E23ADB2076FAFDCE411082795F7FF9C192C9558C903",
		"8624600022D751E11A1F20AC125C02C741B0C7540964B053318FB841A8B251EC179C66D720136828",
		"560D07E3B9C7159700C07CA496EE1465B8800002D2F042010106C71741458570D41D04226CFBD125",
		"C714F34C9ED699F1F1B5A29581547D80C8244A7E2615E080A58984714F54C24E8A0D17022918E327",
		"B9BAE6870127292010BD8AB6DBD2D2064346450579F0CA5ACFCB6CC9D43621D7051FAB8E6C7A356A",
		"9E90B0D4BA3897538D530D02A3B7DEDA6868832F2E7CF9E09BEC2AAA0D097DF7A3D29F8B185D8336",
		"63190C6A425FDA69F99E9638604079ED41A511A4B328200115092F63521C1D4205CC2582F44907A0",
		"99C865025957BC5813F79179EA4832054CF6EEFBEFC0072FFC1141000021FF0B4D414347436F6E20",
		"04031039000000015772697474656E20627920474946436F6E76657274657220322E342E33206F66",
		"204D6F6E6461792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_19 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE0000002200003200000000A50000A50000A50000A50000A50000A50000A50000A5",
		"0000A5000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C3A9F1D",
		"8B0D5A4C9D4EA4C4E753203528024A87954AD96CB1534D40B2914E825AA3613BD90614DB9742D953",
		"4E09011F242A5447367988670D058D018E01918D8E938C11242405739B578554290D111A09930515",
		"579B8C5A911F2210201F5A222626AF20090F20920D599A5F0357671D1429898643676D9B050A9159",
		"0E0D8F92908FBEA9D77629C74F6667767370706036026788652C422C23995B812353DB4A1C2A14B9",
		"20110DF2FBFCFDFEFF00030AEC5006D30705848E2140A0C013820123F260D3930A15A36B74AE491C",
		"F12586187E125D2CEA12ED5186695B4088F870520B974C8C7CD59983A84283112DD22DB1A140C12D",
		"FF1020144868636E0E094991329C8C1420C3962D493F54CB0253DC958802CD842947714E1F8C34C1",
		"A522578E828B2F35FA0C1442410E426612785151815380B6B578F3EADDCBB72F5E0A0AFEA8E06088",
		"0F26982354E86CC28041A2337CC49029368C028A13BA70D530948C84865C9804B8488462939E0174",
		"F0801500914EB0135FD8614A4062B13C167706289AB9A9B40312117C1598668D26EA3E67EEFAB362",
		"4700EC984C4B454A104810D36A0DA88F9D53CED0E4A22FCAC69833FC7AAB57813E40580F4B50AE48",
		"8CB057144D82580A04F18EA420718F3A2CDA74FCB14900BA1470DE03B1A807C13DEEDC825D711135",
		"7082724F70B09F1A359050CE73D74C12FF1D52D70D6700232428400A6C883420400314FA93876AA8",
		"3442CA1C1594D208268D7082CA092FBC3080001750A0DB40DDCCC1111D5DA5725C6E5708504E030E",
		"54408A039720A4005EB8DD440126889C20C113281490425025E6554C2ED42560A55F203D00189B70",
		"C629E79C74D669E79D78E649852284E9D9440A1560109727544C66813EFC90611B1129706560040F",
		"6C068519B391420228232C3A441966BC60438F349CA19B396784E1C2A9351070C20306805026142C",
		"D883C928112810510B36D460030A5C9D254769D7DCA1A257C040E6420B14C83608A150B881891C2D",
		"4404CF4C275C56D115167132070D34CCA19B001CD5C7080AFF34FAC709EC8045FF5E03A575F15B71",
		"E06C4987737D02748793C85D04530106C0078954FB7A010E6C76D8D022483FB2D6000D0E84389C74",
		"D1959249010978BBC8170099110C45323E128823D529F5C85339F262CD576834408FA64EA4C04067",
		"5764224D002E4902147B5C0420428193C054113108A4D0420B071B71420559D0EC050531789A8A16",
		"D4B502542C0980B0DE2B1984F08A33D93D12D6C68BA4204512A09030B53B2772380775ACCC82A016",
		"56B317C9D6D590A2491E2BDAB1180B2C1351434FED2560C050906DE24201200CF781D50A9C148B08",
		"3B43A5D294EB2E4CCE5708EC592980482A4C9E87ACCC7C9D521F378290C09BAC88A877CA7180C008",
		"6164B44C35B447FCBAB0D7249EF55C1FD9F4BD4DA95F381BDD6F32534209EA18C131000A06033441",
		"0CC282F19C26DC3A006F71144526804736C010111F01D143CE2202EB21604633A136AC57522C0243",
		"05156C490C918B38C7CB17A944F61557A4A620CA071A18454F46E002BC14A960A432073A92508302",
		"6820001A485CD1ECC58B1414E02C2E18C1959EC00E15D86A04F522126CAA4631126470834F60C160",
		"FA528F13182801C141A19F8E20071B0C6582331CC20550B0C21CFAF087400C225F82000021FF0B4D",
		"414347436F6E2004031039000000015772697474656E20627920474946436F6E7665727465722032",
		"2E342E33206F66204D6F6E6461792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_20 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE0000002200003200000000B90000B90000B90000B90000B90000B90000B90000B9",
		"0000B9000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C3A9914",
		"C92641523D930A67CAC61DD9048DC689D20087B9A701D7356A91D5D7A28B9415A60414DBE9C40D37",
		"0A7F01820584840105010D89052409807E7E3614152362292C7143790D24036188831F21100A2009",
		"091F10261020A61F2220838F61343603645E5712150AA509AD24240D5C649081860E8A09B28587C2",
		"C365140231B79E611424994C297936610327900302B6C49849020A2B24E7DA4F140A8D1FA7241CEE",
		"F7F8F9FAFBFCFDDA294458A418B022450569243604006123130B3D36508130A0A0C18811145070A1",
		"01A998376F7F1C3820312B92472E00FD49B2E5A7808165820E21FAF00104AC0F825A6538D4C01121",
		"FF68E1F85C6221B05D92132440CC2B04C85BB8300170CE841533C3879D38332C2BD44818C89346F5",
		"3D9474269CB09290BE5E0B27801C89084A692282E3AF080B4E290ACCA9D0C05EDDBF1C085040F0B7",
		"B0E1C3FCEEE0A98798DF586FF31E54F0DBD8080B0B5CDA72F9E24D1A85361706046BA4206C932DD8",
		"1C05B37849088B6E95688D09F394B66D637CC6E0E386B49402BD6DFA1833F653514C927F4647D2D3",
		"80426B7E14C6DDDAA30894F543A702EC8C5AB3D09F04C6920998830D818B8699B67C242B6C107741",
		"0920C897AF5D3E4C403E1BA8F1E84E2048492F08F71D4CAF98F0C029A9C847134E357DF0533CC8D1",
		"36C2734E2005022BBF68104153028403FF080932C124CB21D8B957DC70FA8D171D7A5A2857830208",
		"5070004B2025679D481534525C49B50DA0064B3518504004F110B60F075EE0A10627650D77110564",
		"9063434A4290E0C0071A68100F65867523064436D0700551950D91020A0E98724A0A2394A98F0D24",
		"D8B001636ED669E79D78E6A9E79EF860A682697C32F19064DE04FA440AC18010410A3518BA448C61",
		"269025082970892737C44892C7002EB031000A0214B00C080D007A98402FA4AAD130C444570617D6",
		"5C144C018DF697020C0D0044A6112960568368B51513063964A040469CE434C789A5EF9070207CC1",
		"606483045C68644B6D1F0DF30C8A6290410617F8D8A0C0036F45B041042440D9FF421F2768C48731",
		"3470441253616CFBD42614DCE39F0471F2E2C267C36093861F2870428290CD9098DC86E0D106121F",
		"23D4936F3E9952F02E23F322A2318922CAD4D223DE38A0D1842AA490420B2A4C1C070B2DFC687144",
		"F0C94488885661271753F3A6C599C5230CF002954F3C46CE53A21EB2DD213589C0CA3C093C00C120",
		"3DF1F4CD1EE0EEE3CD79DE1812532A4A8D3ADF03A680F0004E329F08651846F667C10BE4F8DC929A",
		"5129DD754DF23DF0805535E547C833EBE51AC7169C04404504D2D4164676714330F68236B502D34E",
		"7BAF36EC19CE5D910227BF34F84145D7FA01A2D151696774C78408F909095E18F385462A2F212E08",
		"0A6497C00618C4C9CCD9271B260C4A4C093B83560330D8020303D10DE6840D23F432CF33B53C5563",
		"E0D8CD6C0822AA3D53F03702F82C09A77C9DB7B20461E5B1075B4F61D7000D159C18C9197E0810E0",
		"4809441081060928D042DAF904E86D37866F028D18CD19C7662401103815004BB100415DEEC2196F",
		"3C830FD6B8461EA444A121DC25570CC98661FCB3193EACC706C6025A1254708B3F95493D1113C608",
		"5C5007471DE158C148C02E34E842231CEB370E0A460D91A080319402041B60D10E8FC0AF1448C005",
		"436402B392C8C4263AD10941000021FF0B4D414347436F6E2004031039000000015772697474656E",
		"20627920474946436F6E76657274657220322E342E33206F66204D6F6E6461792C204D6179203235",
		"2C2031393938003B"
	};

	private static final String[] TURTLE_21 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE0000002200003200000000D30000D30000D30000D30000D30000D30000D30000D3",
		"0000D3000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C3A8B29",
		"9BED94086814ACA7964979491B0DDBA04119501A28B01A2CCD6E893612A9F0011914230AA525B0AD",
		"C1340D0E24060505018688018B866A05246A670D27521D291D6E6F4229617D7E608701091F0F2020",
		"09A3262610A51975A78D72056C6794362C994C2C7EA686869062A08F88BE8C898705A801726A2790",
		"657E771B1ABE149A49132D52D0270D237F61DBB9D7E4492C23240A0AA809201CE5F0F1F2F3E45129",
		"0353E3F49A142C142914C4B838D3C2C50817B2C8EC3B1245400B3461065002D5604EA3477FA47CA3",
		"04A31982149774016CA06048177C67FC2422962841061022407C58F400C2A204B350356A60068C00",
		"FF7D6F029EE8E288CEA20F22444050902180AB998B14F99AB5E6048C3E021AD4F0668D0940750F5E",
		"2500F6692218748748246810484D593D0DFA7485436205C885E6A6B49003A9C13BBC78590C1CF117",
		"B06125210F1BF667474105098AE171E090A2020189554090981BF9C82E1B01B7F5942245C0B7872E",
		"3A13F167E31BB8D76C048C017362008272FE28184AB17980000A156CA4F9D3EDCFA20C19663DC289",
		"F38F83490210562091A2450CCE4728B8502007840207232C5818E18742F18AC790296B1AE0C3CCA9",
		"38A90E90C8669E0A4F921C1430C03CAA2810100408C1071924508A310D50354937AD61774D140DB8",
		"F0452838B5176058A32895947BFEFDA2FF0027E7859147192D041547031558F0C91A018010952BA3",
		"203355328CF4559E195FC4304B040A0CE4153AACB493C01D0A9496D51AA12C538003D3CD41511860",
		"C0451A2E606810C0061F3EC1C20A71C8714249268526892765F1B4861E36F491D81136DC161940A1",
		"7D21052567A8460E0B816836164276DE19C603E94800669F84166AE8A188D2A3E208841596E81336",
		"2880015F8F6A7906091A24A04003953281000569602ACA099D7A46620BA08E21C75840111A05695F",
		"F0A491378C9E11C39A7DB2608117A499858FAA62A439C00887BE4AC250918451DCB267A4049A9DBA",
		"9687461950AAD117389F7C62860A1CB09042AB88554441919764D1D0147EB896FFD5277320E22409",
		"1124F38724ADB58080042AA480C00BE00A1190020F1003890B7CA080AEC113F5156F82A11CC2B0B5",
		"541D59516B64F6D1EF2E649812413BCCF4519103C3597B533155B4F7014E168D45951FE83AB8C467",
		"243C105016A0E1435A03631123CA2125BB3753023155915373B2BA7642896FB83944CDF309577287",
		"87B8A7944C554C6D8C7260182C09A7F18096E624604855802B372105C12AA6B4271322CFF1A56019",
		"FFC453CBCDC81CA59402A7B4A7542954FBFC4B732142D7463D7E90309B21275701B4090FA0E2734C",
		"273B1580010C17E0DA7003C41003037A50E0A813FEBC30C00B5E1C9E4CCF303172D3EA5239495B40",
		"310C4BC108D37132D5E812519080487353549B20234FAF74B592B3A0705E0C520C002F7F57C4FD32",
		"CE8E9BA25938B14EA5F392B210F3E48260D80CAF8B1F680042010F341147290ACCF4810212845351",
		"88C330C256050A2A4B46B0522466510A2044FAC474EAC01B2448E00217A840233762D71A5012093D",
		"A0042808184141FA95846E41864DA2338350FA3291669D40006AEAD4B16E26AB39D9A202147CD4CD",
		"282007827DAE5445E8C63A34E502AEC1904D3420C1F83E508014BCF0864288D4080A0082086CC006",
		"40AC600A0251991F2691081C50C113A748C52A3E21080021FF0B4D414347436F6E20040310390000",
		"00015772697474656E20627920474946436F6E76657274657220322E342E33206F66204D6F6E6461",
		"792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_22 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE0000002200003200000000E80000E80000E80000E80000E80000E80000E80000E8",
		"0000E8000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C3A3914",
		"0AE1344A243E249673CBB5BC6C8DD3A0411150C28D866D9C5693592C05C7D66A45B55C246B7D6AA7",
		"490D050581820501871F1F8505098569026669032D795F030A0D78146B6602608C09010519191F87",
		"011F202610100F201F57A683240A2E7D28692F2F30670D2D031C79422C2F9E367D7EA882A18785A2",
		"CF690527C803D336781D150D0A29365C51200A20205680082E1D51366B6765608069232427EB15C2",
		"F7F84223DBE21111DEF9020A1C4850188501E41490B05790201C750D06AC4B63A74503120DB774A3",
		"A0EDA09B060E481410D92081C8047EC0FCB1B18905856F141E84FA1041E1184FF2C8FC11D4EC54A2",
		"FF57A1444020274AD0CE94EB28C00998420DBD4F861A595135141629538742190A00AF8D8D18EB48",
		"24D800AB9109272E1B9063F441418175363CF9294922EB201A68DA5038E1095BC6807B6C9010A940",
		"4283607F130FEB406084E2C7C35EA89BE617F2C0C05B4150406C191F0B0175288CA0E0627003339D",
		"99045E37E00CBD78698EADAB9C3AF2A6D6BDE6EA0E732CB600170852B448A1C24B8A861462A4C1DB",
		"5B655D520102F971E4C7DAB57BC40A881E5D06C5B16978D3A0201121EB3365A82024C81000A5D846",
		"83BC6F52A3F11D08B10F1C445167ECE2D645828CC34A2BA97C20C22B828444D25C2CD510053A4F14",
		"F1CB1A1FF1D4DE071088038B81AC1C68FF0A7B060C92921A02F0D25A01350987CF2624ECD3062888",
		"1488DE7F8598160F2012D9A0000522A1028841179117DB5E840C3648518C38E0400546A5B41F4B47",
		"CCC199136790E08A15561836516C8FA834571FDAE5A8146436B8E0421A2F1DC1800A70C1F5945323",
		"8CD0C294B509F38E5B09A480519DF9B0A0864283EDC9E7A084165A5B0A2CA4100C9D865219130624",
		"E8D9A8302C9C7045020F90C0E8A448C055804CB034C0E9121C00D0C208130DA68003B48D9A4E9907",
		"8526DA1916B8DAC16A718121518E9E6C72DCA8DDACB18E000D8C70022E6D9070065C9BD6E6854A14",
		"7A292D32BDA901431A7AA6201C407FED41EC6E17256BA41F639030AC0BA8D261FFC3AF0D59608336",
		"C7CE17488F23CD2BD58263F4F214BB0DB934AC4E0007728A2100A66192516BB041063E1C70A0A80D",
		"2A5080A80D2830F71A20212169458C580D028A88EDF4D242AB494451D748809026EF6ECF60894A22",
		"A880F0C0C04D8A07C609248C8C968E759193C07D39A34AC627752119552A43BD124006E3C8528003",
		"8C9044219A4931411A260A60224107A8EE4AE429609F92C080AB28F041290716A28049BABD300203",
		"0E16BB041CFB7086EB8BED8972852B4471E80ACC8784481D3C319851470591EED8AC1207092011B1",
		"04C7C8CA9597025E54515D1DB4C91A4F63B9B041EBB4F05A66CC70084233ECD9C548330D204B8644",
		"FE24F28106DC7896C494442FC6B2F4D219689577CBF5A621C6317D75408203792A3002C973C72580",
		"BBC227DB8C854F8B45703468425235112C0C90C266402209083D2B1BB555510D54003597641C0425",
		"72056CF0F22BB498CB92B97E98366EC0ED17FF979E32335B00C8010814786700A83A816B24820C35",
		"A8444C96415CD6C40188CAA4E00BEC90CBD09E321105F08903CC13820222F6A489F06804DA589CAD",
		"D271113C9D4985B6A2400DB0640539AC500F82C1C4381EE0C11B26018478D911847C482A221AF188",
		"48DC4210000021FF0B4D414347436F6E2004031039000000015772697474656E2062792047494643",
		"6F6E76657274657220322E342E33206F66204D6F6E6461792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_23 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE000000220000320000000084000084000084000084000084000084000084000084",
		"000084000021F9040100001D002C00000000440044004506FFC08E70482C1A8F45160BC051209FD0",
		"A8748AA00C060D92B6F0A24CBFE0A4CD26B035AE9486BA71B29D4E0DF7DA364A91C61456186C81A3",
		"D4666B050183831F1F09058A592489058058032714362C295F2C360984010A24230D236371709B01",
		"9CA70109871026100F2020A901066B6B2703943614091B1F1A20051C7B4394272E708209A683A985",
		"8E8FB70D6995C4516920AB201F061B246DC96D1469BA5A736D0223D6EBEC1D36240A0DB11B0DEDF6",
		"F7F8F9490D091A090F2BF409546246D7980164F040B924B0488A74B6D6905844224284020926DA82",
		"1328CE307D9950FCE948838433531FB67D08F0E1C183958A1C647914E8D1001816D4904831C20592",
		"FF4CF01A2818A1AE430A1B0877096A7648C4AB07CB02644044C8D1C46827408913C28246830744D9",
		"B178F162449933661E297A644A5534846BAC546B28E49D82BB0A406839328C0307251D08D2518031",
		"851DBA8813D3B560C39304052D48E8519C2F9319051B02E8F542995D0A0A2E28B468416104E97A9D",
		"C15826930BE91A50C6268506D520858A140818334C6C0185575C6C220ADA384721E54C69A6C13951",
		"C064AA066B972E423B227983C9FA8ECAD1A9E6E25A96555BCA3AB55691C649906A8CA3F0314A131B",
		"29A68D4BAA06CBCC66CBB441D8FF6A65D30F8B78C2060971C4904B0C1514B08102B92041493CBFFC",
		"53400AB9BCA0940D184555C054B164C3FF12084EA944DE29E6D882103C84B1A4800D5228C1593102",
		"08901416F86D9280084EADC449338324005D70039431D7101CECC68E3194D8A7068FE525E2400509",
		"AAF1461608BD4817051962C4C618DBBD26DC2868D1801D6227C003952A7A35E0824F1468771042CB",
		"8D42C1271534D09E626DA6A000606164224F230580A0426AF868610609A211AAE8A28C12BADA1B56",
		"36BAC738128060C05DA8493AA919A7E82599A660504800019A34424205A05251820B038C30801669",
		"58902A1499B48046032D54D7C0A0B33AB41A20D36C34460B4470D0C224AE469AD86AAE7DE3951AD6",
		"C501480C2DA42041642D08A0EC6263A0079770F50547020D6FC498C2988A5960FF61B0E06A01685C",
		"7058B74B6A969DF38803E6FD188D228E40AB1420E8EA83140534041288338414D264C2D1EC22C908",
		"01DF43096CB42DC22F46A71CB25206CD41A7D197098D71A7359928D9510423A2B4D23F108C470874",
		"197DD9530DAD3911EA082AD850B01C750A92702A20EC07E22937A6944800F0F878C2B34809104A16",
		"0AB8C0EB4F3034A702813154E70681186AC1F05A0904DDCA7E2F21A2637952522063200E68508809",
		"4F50909766094440820B16B4404A1B13A9A24A0644E7E7D47E216C2315791D437A86850680A0C12F",
		"9E8C3C849F7791000205A852F042906A2FDD23D11FB40CCB321AEBD856C2711018C72EF1151635AD",
		"16BE4302116AE351CC13799B8068022C232ABC96EA6C500029112C0CD0824F4702328A211EB29492",
		"5B9FA3BEC6155C467C4F26658EF082208DC882513301E06B7734B68C33AF62DABD908C441529928A",
		"46CF964FC9B620C91E01249408B74573E44327D79088995378087387317C23107002931A68330614",
		"74E604FE38447E3C61031490E10D4A315F41120807170885327680854A7C842AE2B5864B283CD44E",
		"AE6533C5D4C913B0DA030BE0E08944ECA55751F053364010810D1809870E1A073CDE813C204E8105",
		"1CB8801197C8C4263A313541000021FF0B4D414347436F6E2004031039000000015772697474656E",
		"20627920474946436F6E76657274657220322E342E33206F66204D6F6E6461792C204D6179203235",
		"2C2031393938003B"
	};

	private static final String[] TURTLE_24 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE00000022000032000000008C00008C00008C00008C00008C00008C00008C00008C",
		"00008C000021F9040100001D002C00000000440044004506FFC08E70482C1A8F428E12C96C3A9FD0",
		"0EA532B03548A4C68015ED7A8D149B98726D98CD56AD99EC68D84E36CA772E4C9D4E8D30FE1A0814",
		"0A01191F1F05190509050D89578B6970162335570A231C735C28363466598A7F9F7F1F1022262610",
		"0F1F200981800558347B0D02365C0AB00F954D61360F050A24725236B30255147F011F09ABC9C919",
		"7D010D09589D66236EC174516424118D636E670327B35714776259152496DA502CDC580D155C5F14",
		"6153C023EEFCFDFEFFF0A4290371A5DEBF834358C411C3B08A9B612E025648A12205891136103261",
		"0143D29A339DBA1D5234ED8F190179C8E00190A283C17F14049039978584013F01560D1A14080484",
		"FF54841261B97632CC001209340CFAE584E39F69245C34906A639C959C37434100F10084B255CFFA",
		"7C0A90C5CA892A193512097325469A332E024430032DE7223305C2D068A9B6888D0809062D03D1C9",
		"854B972D5A60496190C535120ABE2A78D9B7B2652F2C5E5020210132891A0A2E6B14738204889C09",
		"1AF015ED8E82D4112328246EA0E2088B14140654A1CC3A05C3706788B2456363840B0A2956B31642",
		"1A8FC3339CEEAA31334E5C36CB2C04ECC193C8012012A040AD42B408C51E2B026814DBCC9B0E851A",
		"03D0C4C77B28B09F045C4530FBE3008B7437328D5041011BA4D65E14C39871020A0D30E807207D7C",
		"200204143EA04060CB3CF35F1E55FD05FF820205684098724DB040C20014C0C65616271C521734CB",
		"E8048208227CF0E22A9D9C85072D45B0706042143C900C7E09E0668305BA39D409841F64302385FB",
		"D565522278CCC40F47620C105A426FECF816495E81725F229B7C948700245A1686003038C71D748A",
		"54A3201E248CB1A5681490871F640F9D40C618285601C76FE015E9C29D95D960DA87AA7CD0C08FB6",
		"6DA20A082074B6DCA59866AAE9A67424F7A7A29C5E728F02182410C18790868A040B787C60402AC0",
		"A81A856F281060832BF2A425AB1329B420C005645C63581369667A4F0B035C93DB00B350F012070D",
		"8C40510B315C8729195AC4610D1AE7C551810A759EE063A6F064D9001CF39D35FFDF1B66C0424E1C",
		"9BFA368B8E68C0099D15D8BA01070298FA66C59767945452228BE826400BD832C69A052FA0942F78",
		"2665711728F792A11B4362F0AB9105289D712B098188D5CA279C2002E73D271957C33DA5FD531570",
		"3630D80A3483D4080DC987C8491D0C0D5CD080020A207C25CFF39C51C05C633D43480215DA481E64",
		"09B4E0480C30B0E18D1F887EA1501A0C9A118158CCF894807ECAE4F4D383269D1BA8188169901316",
		"DAC80B9CD1F725930A85A67465634E6853298EA2564CE369173EA6F85B0334CC373232118E4D218D",
		"5E8925B21B44C99410B22D40F1CB5117C5109B181829FA941F8250EA55603BED4D1E59D5A0656D17",
		"9B8D170116F718C5E2D61F0630236185918BC93774CBDA502C1414FCB2CCA935586254827B8057B7",
		"8434DA05087977E11B46AAC42A0A8C020D10914292BFB552761F6189D5DF80D2E5C6A33F5C684C04",
		"AB569C0827E33393BC49C7C4618A712C8A7862C6C01E7344CCDAB19C17D8A002E718C62EE0F42F80",
		"C1811C2DE81E6B14A22744A0E50C779889A08873052C48A00504AC0C0B1AD02810682002BFB00101",
		"7681318CDD0374BF408A0A2428C2D2280002A8FBC067BCA0896514E0030E98CC654C6421A0FDE275",
		"4F38410C15A028246A44210A7307072653010E102084BBCAA216B7C8C52E7AF10941000021FF0B4D",
		"414347436F6E2004031039000000015772697474656E20627920474946436F6E7665727465722032",
		"2E342E33206F66204D6F6E6461792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_25 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE00000022000032000000006B00006B00006B00006B00006B00006B00006B00006B",
		"00006B000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C766C02",
		"C2A051208D1A2D64AA35A2BC52CE309362B30D04A886BA7C6A50D66E755BDD3889EF448BADDD2EF8",
		"031F0501802083867E6A0924056D692F310254111F1F2024247844647C346A24017E090995101026",
		"26102095098382A0740203366099429E9797235C6664288B54A082949519AD0D09150E05746E0214",
		"B57865022F2F6D14277BCCD7B236727B148B092E080AD04D290D1120EB1F8A0DE6F0F1F2F34A162F",
		"140F1B0906200A360A2CE8091CC282820B0A6FA0EC61D340808D172E5290B01170E01216666C2464",
		"962E428265053E225AD6200D9D13065BD4684042410B5AE63052A051460D4857198605C8302A0184",
		"FF0FC14ECE214391C4CA1429106462C1A2463735531619DA59691D08513BA70A225907A10D732928",
		"282001A20188AF425874E156D3A69FADC05A9158D36DA2464C168DB068002A81250C13518CD09822",
		"C6CB061509A2A8A14041BF141538E42DC274B2E521124F90D05CEEB2C03620340448A0A06462CF99",
		"38A4E8728182AE06294E73B884744458D448BA9161C951161D5F65349EC69D22F8948D9EB872DC3D",
		"80CC44A2661AA8181E0F46278E6FA9BC1D49E2A34D3974F48C7051C06381CEF052546B206BCE5643",
		"A340942A142C40CBB9759A3BA421AA8086C6CF98F3821AD7D021C847ACC8271F048E7D50CC2006F8",
		"31170A9AD5419140C5C1E1D67653C597FF0A50A3FDE1C70927C8E2D03C668CD7421903F8E60A2B21",
		"7D208229A9B0024A2B88D071CD2CE9754709692420F01051644CF1872A36F2B4D52FD8C952833C65",
		"28805005E8A5850D4B313C652033556848971A15B0449C0015444052709EF0A25F191414D74D63A2",
		"4440024C79D950C88FED9CA79413C53D570305045CB6D7038D9526016E8826AAE8A24828301DA34B",
		"756183040990438264908651900D3EC67749A66114D746779714008312055197A80A14546003012E",
		"347005A64460041B027AA88A1B022A6031C25A0E95A114056422908243742E9A4273C169E4461B5A",
		"4AE390AEA8B19061035A36C747747474B3509B895ACB264B2472C4CC14D89281FF101DB1C54601B5",
		"F12854A20073DC1252384CCAE14C897B0C50C308CDBDD100ADF16C522E1DF82990E31FFD29E7CC1A",
		"314C795F80F130F042270237A0CC2B7C6D371215EED47B823314C0E0C0471FBC03E515E049551F25",
		"22B06263010EB0E40E6F1A99A100250B80605BBC57CC8142276F8DB2958C0F3CB08E482C71454696",
		"A48D101252598055C34A069A7920202250A20A509414630015F8D1210BC529B88016342CBC309827",
		"E9F4C54A060B960281D24065F007C7DE62631114D8D6CB57887EC8E893DDAA1CA248D944C19B0951",
		"827FB2F768ED88320C88455341601DC98235181B73D5E5CA1F3326FDF58DAE705C871BCDC963ADAC",
		"2DC0D0109A36F861E0C0545E8B302320EF2D72A61B64C4C3C2007FC879D00016C492314B836035C8",
		"8310D2DC409828D1E57813363C50403B961C5A1C3790FB5E78211C1E1847F5D88C40F05265E03296",
		"267B9CD082B96A906766036375C9118936C42A4F0A7A48DB1144D5A2A7744339160ACE25B0D51012",
		"B8C08103C3CD0B90872D8544227071F04A05DB2711A6A18653E170071B06808D66CD22312AA04101",
		"40B0C2023C406593494148FC0202038CC50D6178C30B15601B12A860327BD940001E90A00420E60E",
		"2C50014002B2BE81708A50859A13A89C60AD3444A48953CCA216B7C8C52EE221080021FF0B4D4143",
		"47436F6E2004031039000000015772697474656E20627920474946436F6E76657274657220322E34",
		"2E33206F66204D6F6E6461792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_26 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE000000220000320000000030000030000030000030000030000030000030000030",
		"000030000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C9548E4E05",
		"4582446D308CA9532B95A2A49AE0F0D0D6A0901B0DC1802C409F4F6617C5FCB40D6894DA60C475B1",
		"C44B29362768850587050D011F018D8E870E0D52246E14023136206405270A14804836152824840D",
		"248D058D1F20101026100F22208C018894276A03275FA04A2C6664949B8D098E1FC821B3B5050E91",
		"350D9B6628BE47360F05200E2027442C29D10324142F6D85E7E7C136EBC20A20051F111B0DD56199",
		"8BC6200935F5FDFEFF0093441B41A202BD800893704831A2E1080A27D89962272058423114CB9CA3",
		"52604A0344B612A143936B4E0D031F346C20A10020384A0D4C7D3C9480518646198CB1BA69E01025",
		"FF14364A45FC73D1CC89A38640164396C0D8A354E846AC03C841850D0454A858A0304040C5686854",
		"D14466EBA3539885229EAAE0620487804491A41040A200C768EC6C74B1D0E1AD353229245C44C282",
		"9702052452F81DCCB831C2B9291E6CD0501384BEC44A52ACB1E198890D0104E234187170088B0615",
		"B0DAF8D499890039EC46534023C054A49848D1B420FA4DC562C682EC44A43132ECCC7340CF7DB6D0",
		"E042830D0924216461E7CC99420A3E1E37DE11D70914D130E9AD9540011F803056C71C6137408643",
		"8B903D7045EB4302442273AFE975115C031432CD84D32CAD58468B7B2141858645837501D1462119",
		"D3917D21B4425648AA14B29A198DD950FFC375A7042061061FB8F2CA030AB8A74A23758984C67E09",
		"D931C2002D08B0CE00BB6C029253386128961B1309D00209FF80B3884A2C752088575EAD13201A77",
		"4189D782EB9CD2D114985563064B271814CA67E27DB61A4C9FA993425CC0D090150904F1C7180588",
		"A41481046820D05A3DE04460200888B979E79F80062AE8A0013590820271111A060B520D20417920",
		"3C10949F4750FA27305592608C0122924083A529B4E0820B03B0A6E83A1FDA05D31154A4508AA581",
		"B640C1281750D0105E142080800D31B8E0AA0D892ABA90430DEC82DB391351A442B080DA20D56AA4",
		"91066421B3AD810E02C1D9811B09BF3D464644B3C504A2463181B51A216438EBFF02342C09329819",
		"E6945B88A6B71CA2DD390314F2420C149060C07D1F10899044669C534004D185D514BDA7A0451B0C",
		"6A14A001321A6459E42027E41B96BD62E1779FBD8815F0042573AC63C0030D7C30C20AA6FA030371",
		"E8AC881332FA3882E129F8E2B85A5F12A4E00267E8BD61437BB6A4B24A2C107C20022389D88B8E5E",
		"8C31F00221BBA0401C7CEE21C3CA225C8B988A5D24D506F460EA957B70828B8860227D173AE2B08D",
		"B0FE73C6C80D449060651080008B2BE4C1E7F41CBC0037C8B15763588C4EC8DCF448D3648C606387",
		"B38107A5D18BEC9978D13E1E6B49DCFD50779D001DAD3896086ABBB288BD0954705BB836A0C0AC3F",
		"0C956AA3CE338184C6754D368B980871B3CD566542A73CD4800B2494791D7CACE08EDFE26E900407",
		"692DF5A3C24C0BCF68C766F012D2344811245285BC60D1D5AEABDD8A41811429AD2498928ED7F690",
		"C6C569589DFC27B0D4803E11705BCD35D73480B211B1B1D18B36839B8C2DC8465063949AF4519E2E",
		"D80914D479C10B28E01BB9948339E2AA92EF624329703C600EA92140F912F2922BE5EF3F07E8C0EB",
		"1455041B7C201BFA00012546C842239C461F0FA809324A53C35F50C21D8749CC0A7B88853C008B88",
		"484CA2129518040021FF0B4D414347436F6E2004031039000000015772697474656E206279204749",
		"46436F6E76657274657220322E342E33206F66204D6F6E6461792C204D61792032352C2031393938",
		"003B"
	};

	private static final String[] TURTLE_27 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE0000002200003200000000CAAB00CA0000CA0000CA0000CA0000CA0000CA0000CA",
		"0000CA000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C92CDA6C98",
		"842101AA266AB6A38D8420659BE07091E2A23C6D0D52638D6EB46D663765CDB6C51AAE6901A46888",
		"99670D7169050D0501058919018C8887846C74667F623068736B8587090120109F26100F22208C09",
		"89097470939460700D286A69019CA81FA422A41F8DA76B73685E362E6BAD5A5D0A24654F66708174",
		"99CF8282CC4E0D0F4F35C5943688562057DAE1E2E3E4E5E67F4F0466C3C36714CC67F1AC42CB6E82",
		"E74ECD28CF24090924054ED1EA45420D9A7DCD6C1878D0E083350AE660D4900690442346870E65E0",
		"F48113A746857C3DC147A4CD1C3516692532E5CF1F2253A8A28D24D961D9137B3662CDEAF4AD0A22",
		"FF43C8D8146486A7CF37070CB5D978B0C79A1F2D702CA88AF785A6381B05483C50C0F5A9D5AF60C5",
		"D89BE80A8185AA618958BAD900C4837F6A20166950A10BDAB449C890C17313CD96340E2249DBB2AC",
		"868B8924B8DA304173551B378FE92482A6CA909C390936ECE273F7EA2B4C6B124448652864E906C8",
		"120C9363E36C67737D3101BCF8E18328515540ECD2A4060561375F2D452BFDF1C337510A1EE8FE48",
		"ABE065E0565FF13345CB388404A4AE7704697ADAEB726CDB245A19E0C3D607E815947F94A042057B",
		"058846170F8CD6148C1F0B6C241F40D6D4640AB8808057E1C0E1020C7804F38E1C9084E402405311",
		"15CF1A7C2470CD35C5C0C0C8370F6433FF445F690CF20E1C284D18D64DD7C091C416A78040020671",
		"E1A554296F7504C2773286810D8E39F6E8E38F400689CE4C3FBAA10002E8F9A5843008B8C0637414",
		"3C80814F06D0A24001311CD10002121019645F599130871331B860A69740C251810D17EC05CF33BF",
		"3DE6454DCBBCE342804F86C30E83733083499FC28C1887300C347041031B7042206C4E92E89D3490",
		"FA02CB25989CE5420C0E68F081060AE4B90D1BA041830A40E35146470C31C0A1C007056820A05CE0",
		"BD62D01A050594893FFD18F24F217F8D68C38D0D6031CF38F244865A69A49E9248605CF907A8A7C5",
		"34035A456A14601C2DD8AA54AA690641DB0A5B7438B0DE211D99775B0210EC62FF4022CEB1E62D25",
		"96FC82554A01ED521B27A5EC124006E4F9D7CC57BD3EB14F461C41E0C9270F18B78BB2FD853422C0",
		"24F806470111A8C4D1C1A17CA22F2A0118020C9AE7D850412C7F55CC7140D8D6F6C122876014C9C3",
		"569579102C8638624A6E096C6C0AA9D298E161637ED191122D19D806CA27D816E04060FBF409D930",
		"8D61010B05016DD248293997029245BD3C118C1A12407D0E165B34DA804035BB8451CB8DF8670F05",
		"00F9B3C7BB2B92D1A863926584F22111F006D9545B24B7878515C0DA4A1A095A4081E236812B5934",
		"7DC2F3E1440C31440101C594164004203C55271A0FA6FACBDFEE486EC44DC43ED009E79D51B5D684",
		"F05880D7524B358034C1A26AD9D0F7957DD880F98F4F58B2840DA75061405042BAF2CD76DFE09EBC",
		"161B280041725CFDFCFC925A2976FDF6DC779F63100021FF0B4D414347436F6E2004031039000000",
		"015772697474656E20627920474946436F6E76657274657220322E342E33206F66204D6F6E646179",
		"2C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_28 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE00000022000032000000003C00003C00003C00003C00003C00003C00003C00003C",
		"00003C000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C9342A384B",
		"566A7A4A35AFD8624A40E98E1A23DBA9D1A0881BE386AD2168A0D6ED827AD0567C0A0992AA05CD2E",
		"6D361465646405240D05890119018D018989880D346B032736567E580C2F6984888A1F2210A426A4",
		"221F8D90726865669A593114636BA08D0919A2222022A88C8F0501879F64232424097711122E7DB0",
		"4214035D145F36020274809EC49780028024CF5829C7C75D2E49368F202009ED0DE2F1F2F3F44D53",
		"F57E2C022418EC011AC28615B1F1E21B20408446B4A020A046973224002648E10C5F0A6967C86C5B",
		"1520811C3C240A085233CB8C0D7C4658D8B0C4854CC80C8A1C2D4A1020D52245C75A75C184722009",
		"FF1A9E481800F6E85682A3AA2091A8D04AA3B5162414F4EC90C2060A426382456AC42E19CD08ACD0",
		"98BCC402DAD507802AC853802840041035884C39684D0631AB07A6A24CD1208202100A4850D44BB8",
		"F0541B085A08286B1809071223143C004CE2C34F0B4E2A2020C1F30820411414B838598FC1142956",
		"DB8C583D4B1B313463BA08F004E8C5080A0E347C580022CC54176C10561B548855584234805EB371",
		"EC84601553559E1860F553C86161133968109855AD06D4194FAD0A884E9A480D6A7E0041EA012A54",
		"010C280ACB85B45E842E0DA94AF6A13F7B1093F5C7C82A1A81D7D97D648C044A2234AD074102BBF8",
		"52D423091023000A125464110DAF6915FF0A600F3C0001081D01B3DD206B8033023308C0838F196E",
		"1012934C34C9144C47898C318C36C7B493C00355A0400F0D06A9A106870B4AD24048AF95C1857842",
		"B000E55AC54066D035D435E0020931D4C21042664CA121611654F68106794077C416211D258104DC",
		"358604793050B0040547B5B35E6012C8390F793564E2E7A084166AE8A1F5D8D025A2CFA440410A2D",
		"B43025A11C9C70420DCF55D1C1A47BA850810D9312818002A0852A4E18122430946E092850400C5A",
		"08D0820BA361B25359E449034654C68C098B35240C40461E869020880A55B574091919A131007581",
		"B8D04505F225A001647BD9400301D2AC868618698C045A8A0A9261810D230890FFC006BB29406A4F",
		"B46A49D227B5B4D164530DC010C30B36E446C276E90A4A8F0BB7AD348CBD328604894B05924107BF",
		"761696C2182720ECD251A01CC52477C80C431C758D59B3A4914B4650E1821E41120028AF592A704F",
		"66606329310CD2D8DF073029F50975118FE71A210A3BD29F08223E707389377ADC867D87ADE18903",
		"8B385200230DBAF3418D45A1278D002FE3C3000A48AEA1722319B0C31E29ECD8C4E0713B11C60921",
		"80A0A07283A34060CA88574BCD4A1A801406C3094406A25F893479A55E2A370273DC3755F82ACF16",
		"C4383014511DA5DDCE800538B0DD55B32C09C808157C81924A238C71D5552AE7428ADDBDFC228C21",
		"729CF179C7200D36D00F42DF9180743036E5ADB82A9F5C72C208292870D47A294830C29F60031E47",
		"30064C5EA2DE88E4845F376B60FA4003DB2390973C2CBCE1DC68963A704848C078A43021B3ED1408",
		"112C84718254890A10C66774CC712F315B0362AA45F6BBCD6D64630D8358026E3C3BC4FD0630A811",
		"A8202A0918011402710D9209805F74A14BA104108C76800005A1AACA0B2850AC080C0301862287C6",
		"8E413F234CCC23C9780B09E272A840700F132844820DAE96B7FE30AD50FFD34244D8E12E123C8051",
		"5860015B8CA800E221F189508CE21383000021FF0B4D414347436F6E200403103900000001577269",
		"7474656E20627920474946436F6E76657274657220322E342E33206F66204D6F6E6461792C204D61",
		"792032352C2031393938003B"
	};

	private static final String[] TURTLE_29 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE0000002200003200000000C2AB00C20000C20000C20000C20000C20000C20000C2",
		"0000C2000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C0E6D29",
		"922D991A51062345CA66723653140988441A17488D56C3C63E35DE701B45D040B9DDEF9340606B90",
		"585E8143297836746F64116F0505018E8E098E050D67937927271482815B789467928E1F22221026",
		"261010201F918F09706F7C58240A9B4C57350370349F8D191F2022ABAB018C8F937D6F6C14232835",
		"B43680B64815286C369903BA86C9790D7A23DF50D2D3811C2F8E201120231C482C36C5AC01C00D29",
		"E5F8F9FAFBFCFD1C0A7E5464AAA0A98880112EE6B08135A28F9C180DC279F890218102172EFA7500",
		"D382428B866EAE786A9080448104092691A185664D9C681A8D5C1B90EC8CC946AC3EE88C44F151B1",
		"FF93DE48F0894984C58B3714F034F2690C65CE00191A1570000BDB1A1B2649A02CA040C20015FAE0",
		"394473F2A481492741A0FC69A901853734EF112D9202CB08122D06506043F32A5F0A9524A04130F7",
		"169B172FC82179CB75CC8AC290234F1B60E185DC251CDC495EA28204180AD238C8DD62436F0D3F1A",
		"12805030659F14051154B34AD98070070E08689EC0BA3B1385B7B01A0898434101089234460850BC",
		"0906857B7B5110C0E27199DB64BAE030A284663789130332F1B93C97828B85DE28955C64899103B4",
		"6DFB646213796F7038248A454AF0E1C103100006108954DDB9350779FD9026077E057C50CC23AC80",
		"F0DF03C01C93922585B446140C346407FF4B53F48000C1283A39B894801554209C5BDF9C4042052E",
		"44E1D93E0CD061D51B11E8D714800FA002420692E4D7521E7CD1A256020FB875803E2F9CE7D222C7",
		"48D28A24EC7D33001F50D045540A28F0A28C556431D21625447253DC6642D4E5C20877094093216E",
		"92A09C7CD7B031029A453C5740041A9020411247A504426A0938500B9E8390D1800214104405091F",
		"28F080710021AA0F829666AAE9A69C6E4A5AA7B6A410E308CC1921907905711A450A15647944696F",
		"D1A2404A29685658A942BC888014509486147123A87181019168A000090360BA09079990306919AF",
		"C440D7352E85B446526B08F0C26F277D508006233C30803E27549080060110BBFF921FA4C142552C",
		"E1D9D1802E93C010C342118C50430AB80A62C35DF27D135138E294B65037F70577C5BD1AC6044FA3",
		"02B860208BF789B91D1967C0E1214D91B130024D0709D78D9891603C95AC6312A92C3FBFCD31921F",
		"406D67B1C5DCC1211E4C8561836D4DEB8909244FC41860802493F8B64661A4BD750899C68C928A2A",
		"3B41C54849267D73901C7389D5477626FD146588ABE824A031948474E05C72ECE6E152276590815A",
		"4F3F408A83C7A46C030D2AD8CA0F3C28B885097C27D1534A2A26949200905289191C1B128F50ABDE",
		"E530F0021EBBA1001F4F250AB8D68324D5E1B732EC1EFB2D1A99E1F35B1E748032F552C000A8964F",
		"29556D333351A004DC607128E4B3E01B04272ECAD31088D0D383C5E0B7177A2CFDF6673E5CC2D137",
		"0DC6102D762852AD770D260501921456095AE549D7539E64FB070628128B81F46D26070D27784949",
		"F1F9B105B3C2BAE129968B2D38D477C271D86C430C1340140B8853A72B1D0F1B7A400A96AEA2A97F",
		"C5A0656EBA121FDC20941778EF1B28B08DA538709763DD69231D60410A3E6383A344A00063D8001A",
		"FA05190E748D15ACA1827A00A49A5768EA05B3FAC04A0E75040A384527ABF9430315708207604D09",
		"36D880A414A000C7714A842C2C0A0B3E62830AAC0C5458CCA2168912040021FF0B4D414347436F6E",
		"2004031039000000015772697474656E20627920474946436F6E76657274657220322E342E33206F",
		"66204D6F6E6461792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_30 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE000000220000320000000032000032000032000032000032000032000032000032",
		"000032000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C95CB25C88",
		"160599B20D060DD2269100915CD3A6B8536950B28E8685E81234DE6F8AC0D640C11B343BC56AB385",
		"C780292E2A237B02240D36270D055C6F059001925C1906060592888B667B2C80807B74676F0E0119",
		"921F191F2020102610102209A78D05241565A3892724245CBF140F281C9F457B287D702492019801",
		"1F222120091F0519908C052356723629C5E02C36346F247B890D27739B67577D7D9EE0F243142A01",
		"115E7F47365CAC0A0F0ABECD1B48B0A0C183088F28188180449F73E91CA6B311C34F960F240A2422",
		"567044850D5EBC3418D1A24503378B4ED07904A911A35E0932C23931C29D8D8446AA9C18B589E507",
		"FF68201E8448F0ECE787000922694497080505120A708AD37362113946CD906AA5B64A81D666CB10",
		"BD1991A8CF0972E648204C31601BD32C982E41FA692BCB9D772F26E074D29642A10102AE0CA8FA6E",
		"40462E111424A8C171EF121B0F462870182F090B1629285476CCB973932A4C14D8F832E04163CF54",
		"2820D85639DE9E1A144CBA30E0219586142B289C46DD41458C6D741651181E074E8C8AFCA22654D0",
		"2085E2A331C5865AA90BCE52B17778FA3958C54F770202FC9EE4B90952D298D91CB47C89FD64A2CD",
		"08A574AB4E2242C648CF58B13A9A35E952E17B0894100B7B9CE1C623CB3033C92A41EDA7A0244B69",
		"27400B5FA4808873BB8D210E16E5D9D2FF8C3301B0020108461195950328CEF4C6262E28304D2F24",
		"D440100C2B1D58400410B684142B2698200208394AF2C64A6678268700348C620788CE9897DF3D69",
		"1437875ABC755747035830925436D8DC655822FAF036440D73BC63C31C03584142212F2852CE1B08",
		"8869840D6409F20D7C438CD6C834F93580A7671C08808A7D9329414104F60134C29F62FAE18D9C90",
		"462AE9A494428A800B0256DA410A2FBC1026114F98E19C4C246428A7002A4881C41E23D460860144",
		"690082022DC419E953120CE7C913830DA7D20831500058041B14A0412F1464CA9B02150852E64C07",
		"BE71C51C14188088049D71A0996645BCB3621F2DA8B4221C670ED0E64D38A590FF05402432A7DBA6",
		"4462278A75BB9015581C36300A0E09032826C916BD14A2121DC960C94836EC25E0C28A8B48F46841",
		"2C08F040192A2549406C02F464061C4905E08B4669F4D2654A86281B0E0CF154111B050BBB095116",
		"882CD5D2CC5C46385C788E71E0429A592AD365460A3A08A1C7325901A66316F581C22630CE15222C",
		"10FCA4604B091C8C2FBA3815AC8BCCCF5083CA070F084DF31D69D2F98685A67ED2870081D1017456",
		"CDFCD40A2C0F88C01F26340BB74B160A2870E306E6A800CE3BF792C28C33D0403DA2DDA740984005",
		"3490539543A335CD8AB02EA4BDC41E1A5BD70C513FD94D1417A87CB81EB96E664442640FCB63C10B",
		"760CB964232632C3DFCA030F8CD878936F827959667E683E0603A39971429638D68209D8B188C078",
		"C7E8A9F3CEA70845DC9E583A4E02E4D45DC2216CBE9E89236D1EE57CC50C35A4F7427628CAA10603",
		"4AB1D712E487F4A633E408C2E3242C0D7414900C917718522250023E4985A22A581ACEF71491A645",
		"98C90628B095A436348243786A0E7320170CCEE0370C20A27DB7F2CB08CCE1A800E58B02AFF34517",
		"BA508D114C0A11F501410394A0AEB871A55092C20A08F0D100CDB1A001D38040DF0A31290BB80048",
		"3294C000246804F191203315D0976732730617604D5380B80C16B7C845D404010021FF0B4D414347",
		"436F6E2004031039000000015772697474656E20627920474946436F6E76657274657220322E342E",
		"33206F66204D6F6E6461792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_31 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE000000220000320000000030000030000030000030000030000030000030000030",
		"000030000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C94C52584E",
		"1BA5314D043E0B904AE56A7ABF8DD1C8663B9D062D1BEAD46037688DC1A011977E9BA441E2B327D1",
		"5B2E0336746D8405877E8701017B010601050124938454143629775F97025484938B918B7C101026",
		"A5221F1990010D05958464656F342414249A47162F830D9DAEA0A1A31FAA8BC606AF84149723B9CE",
		"4264036E0D83BDB236BCD483937E1CCFDF422C0D93112035504A141109240F1234E8E0F2F3F4DF1C",
		"1C99F5DF2C022963E82DE65063636380000A0206382031C29B3E220A3AF821E16000820E2C7AF929",
		"64E85501760A481C6A8082D0894E36E27DB36061480A2E2A28089A52894404578A4094B2428C58A8",
		"FF432395B53941A285823D09D82D73A8E99A1C3A0544024D7AE503A30420403C806095D5226DB254",
		"3E0C7782028D01144AD221714C11A46474A4D8C234F60B0A29026CA14D4866590A18234466BDAA20",
		"62DD250C063CC0AAE0840DA687234B96ACC00689C28846B0103B99C88814CB8E5CAAD1C202951A11",
		"367C2CF0B8B3111614103094158B0E85642F2E85758DD10819656505BA82456DDC08C774EB897900",
		"C22AB711602FD58C34E99003A0C082B25966071CA6CA640810B0D1A2CD20416C48B0C3B9684F5611",
		"0156BD8D54BB8E0D310D5CA450EFC7450B4D2A1CA4DB5D9528424C021F25A0C0030F88D09C318854",
		"505F4124A4B060486348E04C0AB23C35FF8E03108E224229CD1D380A2B6C6DB45177FA64D44B716F",
		"D0D19E159128880A7C19D07888635408225946D410380E548AB8F5111F08BA100B72CDF0C60C1DD3",
		"4C8220541B4169902C17F0F69A2CC825248834748C815682111C358E964464D4420C45FD978F1153",
		"F091556115A239C42D0580C058034B40C1C29B761271CB142F7016E8A18826AAE8A24388D3423F0A",
		"DCC3E8122D2C73D906018010521E906949012F867640460B35544082011A58A54103A0851A193F08",
		"84D628461C4A018300E5A9A74003E71C2A8E0A15A4049B1B675462430CB224E0AA6B0854E04F1983",
		"C834054D505E22880036244A1B1D820CA2D6407151EB18059251C0614B43B840FF93035042B59648",
		"EBD52607B6CB6A52410547E9B9910AD7CC32C75409AC354E2294282353B6F2E4A1C03A06B0C31074",
		"2FC6011582EB0165B12892404914B6E47E930706019F40435ACB8C30C708344595885B8B34976391",
		"905432850DB5A4D0406191F6A6848B4F64344005958E41ED5A232982D5563E1983602470D926CB2D",
		"12A4504004A96D40810A77A4E082001CD7044A4F59811DB6313FBDF2A26381BDA2A7562454908B74",
		"3255898828561153CADDA9289D400415A8757618523D70CB37057147B457AA7C00C1882426CDF475",
		"511162725F9B8106E8DBBF55729D305664955552492940B65B24649E5C3D3E47399CC58CB072F456",
		"791BED8714DDD6EBCACC2570D41741D991280E41894C373D85207C4666412F14F038F04F085ED17A",
		"A6A1C4826D0C9D8EC5020C9F58F606CB6413938C766D74EC5A5AC67EF2112452354D0DB62C6A6901",
		"B6249114B149DCC645866316059A110D6AB950C653D76288005ED009916060126E3B540AEE72BF82",
		"44035A64E84404F29494E6342081685213054660816E5D62332E6900820EF4814DE94F00D09944F5",
		"8A700B3E3C804EB6ABCB1602508ECAACB0511A4909A25CD1B29BB9A00B935AC2321E600305A82B88",
		"5E90820E91C8C426DE21080021FF0B4D414347436F6E2004031039000000015772697474656E2062",
		"7920474946436F6E76657274657220322E342E33206F66204D6F6E6461792C204D61792032352C20",
		"31393938003B"
	};

	private static final String[] TURTLE_32 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE00000022000032000000004D00004D00004D00004D00004D00004D00004D00004D",
		"00004D000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C0E5914",
		"4AA35148683E8114A935727A9B5A49C5469E46A5035B83361D48A7A7938DF5FD726C24C7C0D5A190",
		"E527530D24240109540589018B01890989540D6A0D146936020D120A09091F09240D2A492969747E",
		"142A2D7E6A6F825305011FB01F1F101026B620B28A05068584939229754CA781539328BD8B558A19",
		"9D9E8BB2A0700D2373C4D9422C3623030D272828926A93642324D76E05B40620EF99DAF21D2C03E8",
		"9C35C3F3FBFCFDFE471CF4FD934721C60B0A47FC9C10F0E2451A01343E081CE885C20B320E90BD9A",
		"4228818202915C9D39C58F900203014028B036C2851B4ADF1A7002992843CA072062D1C2D28859A2",
		"FF0A70D41C3BC6C6C6C43ADC5E9621B1AC11276920121802F10057CE4789124D92D38062110B0246",
		"700D44828DA24506A83402896C15990108BC3A614143809F11950608B04181CE131B09723E78A040",
		"0142B94B56764AA9C028E2C790FB49290C4B813D0A7C2267B381C285140A0E3AADACE15773111B2A",
		"489CD00705946A57300E0AB074D8741112155CEC95934CD0318D6D07A439FA98432B57541EB982C4",
		"8B1A384AAB14102291A201824F0A2A50E89284458D4D9E088DC0FB5B50569A8CDEE5B489152B7229",
		"5CA7135E59FB48260C876CD02060A3450B1B27383748044CB1038208BAC4A2004E22D0D253004C21",
		"478924F535C1C01123D8D5C000E511B2FF562C19D4624B83862CC6C8896DCCE6D83F299031891489",
		"14C8CC89298100812EBB64E54A1A037020173734A080067205C2F2E02220401294147CD9E6C70040",
		"01370809CA552389460548074A05B60DE1470DE350B31719BBB9F8023AEC78F20161A2742944146E",
		"9040818F4BC02856855DB290023B0944F0400524B8290F37869D6081A08826AAE8A28C76908E5DD4",
		"055A5AA347ACE482021B707285020870416911490D50C305BD5CC18E03DC21BAA2112DD6D0420C27",
		"B8604D9F36D089A80B29E04A440AFFFD77E50B02C0E062938952D0428B29B0C00D32ACB472C22995",
		"F035A969512090C2355C5D2340792320778D8BD34646825E27742B5C30831440FF82AC8204F32C49",
		"A6A9A0461EC841E2D174541002521C1BC2894D113EB2C001029935A1AC0D87BEC9D1376A64355301",
		"0E9C672484D404E3E2002D542001AE0938A0920204D80A90BCF7B65B461A8270A29C9181F1040B2F",
		"F56AD411539C18C625122C349002081A18F00928DF4A49CD59814585C5538FF4A49C50C8B8712953",
		"857595C408157442E52504747BF224452A4295089DB49C13338F2450010D34F8C12C1951244B5C11",
		"1490D0D8B715B440095F44CA3D8B27B33C608B2D1FD8C44C0311ABCB1132E12A6151697418E7591A",
		"28701549212FD728C22D0846130B234616C01BDBFEAC820905C730C5797AEF40A072027E07FEA174",
		"0046416CE897E02DD548211F4A95748D7E2368939266449BF83C2E4ADEAD2B8BCC240D04244A2315",
		"5BE47423578B9CB9D2B0873E3D9FBCCAD6ECDBC6BF3F0A3B09CACE9D65C859C0476FC301D49E7938",
		"1B1CD14C93E1666C58C6142243B6EC95C7B452CE980310DF20B013154AD8060A65084401AE51090A",
		"CC6628368052048A168B9FB9C92DC33A17E99EA5AC1615A2131F330CA25AE48211A8822F6FA3071E",
		"36209D07C443502910807FD2913F24502005385454CE083141100CE2534BC8192C40801B09080088",
		"4A60800D1400822546018941E48B0DA048C52A1E21080021FF0B4D414347436F6E20040310390000",
		"00015772697474656E20627920474946436F6E76657274657220322E342E33206F66204D6F6E6461",
		"792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_33 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE00000022000032000000003B00003B00003B00003B00003B00003B00003B00003B",
		"00003B000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C96C922829",
		"C5A6F0592448A35673CBEDB00696119133209D288D7421DDA0A0D3B6B88DD2AD2353B60689AD4E24",
		"0A050E80010606010587886B7C6C7114237B35340D364C0A14201F06200559232E0270277A24110D",
		"881FA9542210262610221F0901011FB4888769036E3624764B72796C806B6BB3C48819B387092009",
		"7C360271041CBFBF2C1436030DA127A30D27362703E4E494712F2F247FCD0F200A97D6F2422965EB",
		"EF7AD5F3FBFCFDFEFF008D50A041C1C20B176E1A685010B00E8B060852B8109707CDB76224F630A2",
		"04AE0C2F1A15468C68E0C2851F3F7B24704951A1410C042CE29C3911EA143244CD20D44A65EB1633",
		"FF8D8BF450A2D36504050936F0CC9138229A0D1434BE91C08593164F1001DA4180650B5084A07258",
		"341492CD468D6D6D38EE99842BAB01621AF3843B454281B36735C62AC1F6748000A3D28826614158",
		"6F973DCD9E287861B8B1E32D0878155080A50505158F7FB9A9D1E08201071B3EACA9901949612479",
		"E20CB011235B8C17614B1369918244D214D9DA240497E69B37381D29ACA620D6080705B83352B830",
		"AFC1889ABAD4A44940D9E6A00019370A9749C2458175F09AEAB386166D9AA97F4ED102112B4086AC",
		"7F140162F33B58712E1C54D870965164368BA31470854D593D0041061F80E08E08B1BC778B03CBEC",
		"118C0DF759A247336B1080C2002DB430FFC746C3F8C4DE56EDF9A16026872492861BD960B6D20B82",
		"0DC1411CBC35B0DA796D0DE2478A011818CB329C68C4C66A15FA539630948C82025D11D0C2E32DB8",
		"F021C06A48FA621803307CB3E41E287C435750D2F4428100025800488209204859690CC421000D79",
		"98E1666049558807091B4440021EB21561C308307497429FFC401200081AE85109A1F2E0910DA390",
		"462AE9A48EA5600207F9753028A54B34004F02096E00820A0A08B029A763A4D540041AA462C00734",
		"14892A1E0318A5DB3D368C87EA108EBAE6D784A7124A018C7652E00D92944C0923388EC8DAD80008",
		"7858DC1B73190B0D1B24EC324738C126A1800BDD76519B04158443A3702820CBFFC8246CB0F82B71",
		"1D1CB747461F9000424608C8A34203293495875FD065041E46F3D1479F1C68D4652F4A50CC93020C",
		"713860911A41DDE4242220A6D1D40919C173EF031430244F648D507C0532B56405A580F3AD91DA48",
		"DECC35C7695C60334436E60EC34E03CB24284B563D5145420592BCD1C670D628C07223BFF1B10820",
		"EF89E0CE7A3EFB11A503823CC3488C5B54008F2CD979F88DB55315B34A26CA80B015571910730B09",
		"B4244B66525D24D5E4070A5072011AEA5E083715B620A8762BB1C8D213940150F6CD8D5D9C20C106",
		"721210521BD1188C6D8A1F20C80A2B402B088B7B050FB99A055D3CD2300B2954D3AFADBD384D458A",
		"3BA69C955D10A0E8DCE45BC0B561470ACE6AEA481B027859F62038E1E2F3CF568FE0321ABB3474A4",
		"397A8CC2964FC4B7255A5036CA8102D7FD28258EAA15A5E1C01E6D256080758C08A380042438C06F",
		"432C184463034B7A791EDC87C4250EC42F0C20E05DB2408EAE00F2B0B4D0601B63E3CDF7A4C1A164",
		"F5021099680665C215106C082037BC79C219A8D40D7861E314580181821AA092D264034E72F08605",
		"5633020A762037366841EF1C1393478C6085BB62821B2031951ACC30872CE018DC12108107E42D5F",
		"3934020BA662954BC02B89476813091E10071742710834BBA2167715040021FF0B4D414347436F6E",
		"2004031039000000015772697474656E20627920474946436F6E76657274657220322E342E33206F",
		"66204D6F6E6461792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_34 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE000000220000320000000027000027000027000027000027000027000027000027",
		"000027000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C9545246A3",
		"00480352A55288A676DB4CD91ADFC1EB35107C1B8D93CD66B6A58E9C5523957024488D0AB7C89A23",
		"5C6B0D0314030D05680578058B0601058E8B918B0D1460686D2C7B463616596B67688609888F1020",
		"095210AA22201F09A801198E8EA182022F2D5C0A361B206036832D2E14C4270378A1240E8F920109",
		"91B18F01D301B5C63627091824AF78999A4C2C0C2F361424143469960D0E140202039F836F462C57",
		"E0F849290D24DC1F11FD6CE41B48B0A0C183088DA8185023CF087E0A34084CA889530B0A43529C40",
		"5328D49787B5426DDC18CADD9A1323FA29E8C76FE24015B5BE242A304ADA879B1F32DCA4268DDD22",
		"FF5A6A80ADD1A4F243807E2D46B4C0167292216A1974B63A2505C4031146A77D4884E6041E1B2AF6",
		"D8B3A084C5276024BF7CA9544BD2333C14D6041C446251820F204028D845319C171A035C34B040E1",
		"5BDFC319CBC5C588B8B1E3260ACC22E0F078204A12113468306AA0809ECAE0E2D670114340DCB36B",
		"0A736101332E017C2C0C0FA190024AC8745F46861A440CDE9A0A0D5CB420A1E05537949435B1A04D",
		"A2DCA7E00D50A0E9C7ADAE21692C4386CCBD865E4112656CA46BFB2AD25DA3519D3D9B96A842BA43",
		"D6181F14F02B94249B101EDC7D004204849DD0CC72D43A6684A549031FACF4C408279C50C9001BD5",
		"5553003BE98417045839631586394102FF1F18CDC9C6840AC4A1028283288CE08219229194084FD3",
		"A082CA2319F477D32BAE58F3C9460A6CF1046D4678911A050E6A874804D408C8DE75C860F30509AD",
		"68308D5E2A88884F0AA7D950037D97E84692216904424898E5301057030980B0413FAA3D86250C66",
		"0840C348D80C60C150A00DC48200175587820D3DE63910058FBCA5800BDE09AAE8A28C36EAE8A390",
		"56C60173FD8CA0400A83449A04897BE118800605B8A469111C50828601066C7647A0A316014C520D",
		"8CC0961D40E683809506A5609A00319416433C160CE2A41BB816B117091B5C914542F63015D220D3",
		"0D324021E1D5BA521E2E70739755285CB9866C9588F9A548D185429F0DB29AF4FF09142E1CC29F02",
		"0FD890DC1E14200093405EA041080A5F16908222FD70F5E1B82D9593520334A0D02638369020C139",
		"0234184832772882087B90C404069DDD1DA4820B1530D824296838B04C92CC40B3880382D4126EA2",
		"050582CC75024B03020431CA92A43311A0314A1A41D9502C388269F7CC8469FE27552B3BCDF20809",
		"D370D49BBCF918E6059D95D44CD3CD216490F403A664E535C6771450645C1380C302326CD117AECF",
		"CF8892D7331FF867820958BD925786937CF8C9D049F4710A4B234C4B49C4D345F00C8D8F7CA08A2A",
		"A7E0F500DE204833604CF26951C3031F18E0CB09045462CCCCF641FD133538157AB308B09407A650",
		"3610E0C216FD24A0D8011E2884FE503CCFD9773AC6854AA3D38C37F97D1A3F78D450E5120A9C2081",
		"04B4C59505316C7134B12133A70C8980D2C027FA0933A5D98A020D306146E0591EDE15D0D847000D",
		"2A8DB428000C2FC440E8D67A2B68D01AF184742EF2E341535D1C809B3508660D2F18C1D1A4F08007",
		"90E0330669D6181A7602E998AA41875343034CC306B5C84648C03901CC28822536C4AA2316FC0A30",
		"E221801142AA5936F853079F601AC0B58A18B222810B22D3AA2594A3010A88405EAEB5AC1E92EA5F",
		"14D2CB08DC604438A0015E5F69E212EC6143295AF18AA30A020021FF0B4D414347436F6E20040310",
		"39000000015772697474656E20627920474946436F6E76657274657220322E342E33206F66204D6F",
		"6E6461792C204D61792032352C2031393938003B"
	};

	private static final String[] TURTLE_35 = {
		"47494638396144004400D50000FFFFFF989800111111222222003232666600659800639696999999",
		"99993396966399CC00636496323200336600CCCC98FFFF996699336496639898CC33333330633030",
		"306330636398CC98CCCC0099CB3399CC66CCCCCCFFCBFF66CB33CBCB33CCCC66FFFF33FFFF654444",
		"44636330DDDDDDFFFFCB100000001000777777888888AAAAAABBBBBB555555666666000010000022",
		"000032000044EE000000220000320000000079AB0079000079000079000079000079000079000079",
		"000079000021F9040100001D002C00000000440044004506FFC08E70482C1A8FC8A472C9542A522A",
		"0A45C06A5AAF4D964DE062C0A4B6F040D0B095B7A7B02DC5EE345CA38A24D14824480A520A4BA4C4",
		"A4551D29270D53368405248A24050D0D0501010593948C15288E8E240D69367C4C038EA18E949319",
		"011010A720A910201F1F7620221F91949984027B9F2924B18A23236599C48F90B59294099401CB91",
		"0D929B0D2F36301694AF20DA200D124A2C022F4C146A362828669934349936EC24E43631036B9FF6",
		"F7471424CD8D2714BBF8020A1C48B0A0401523146CD090E08306470AF0D97820019E412283CC5070",
		"84421AC7629C0811DB18320CC0802C5A8C40908282C834A4EC3063952000AC48CE20E98CA4281DB9",
		"FF01015330E0D081DC89502710319A14A956810FB3B4D54AA08D16CE3B23C390A868274185884958",
		"0C5060E70C8511295D6C24538864031266CC3472C42ED19D46EC1A08485AAF088B14812EF26121C1",
		"460B182E16A54320B8B1E3C790234B668240808D0020347C7051462007152A1C176A70C1458D4406",
		"3EB0016C85438AC405A82AF8D7388500CB0206A411204513A70603A4E80353AEC10812891BA013AC",
		"C251184784CC70D6B42F91C7762E7D96A14070009B727C6DB86DE48069B348FC20E12C503EC0A634",
		"C1F782154802C6A1624CCD7FB00A352AB20CB1D82209316A1494C221BB9122CD24FB7DF0D42CA974",
		"05C20326D0020908054C275E18983086FF050B1438B009059C99014F3A24D0E18824E7A157937A37",
		"35A5222E27C4968001016400421E271951C3031F180082782804430F872035206226EAD9D2A40173",
		"E146C20861BCB0A2361B1440A5131454D4E3102D81E7084964C4D5000D2265E74F3CE44CB6040B0C",
		"1C18069B358471DB08160C109A9B03A5745C040968E000273604C6A73D0A30104F5F8736EAE8A390",
		"462AE9A493B16028A5DF98F0D6082020F00077982281801ECB681600431504540809043806CF260E",
		"1810810620E8F109072C142068AD24D4D0186F358C095C3996CDD9C611795490182C76B8341F4120",
		"A220C50B02C0F0C718D8DDD6D61696D5C9DB1B05289047B09E08A64592474127FF92BA2341B70549",
		"B6D666DC462E39570C23A43C92A4BB260DD4126B423437265B31E16BDE3E2F1A838B6F695CFAC91A",
		"2974160621D36592E2529340A9537E0554404A3F6A58062A3E67A950910B08F6A6702931E6889E7A",
		"4A46A0DE5C48317A0F0BCD55502F31932C68132C404BF54106CC241380027315124641CF39B79122",
		"74A83761360FB4625524199C32093122154A100724A5B3F23EA83CD04C2A2654D8CC07555FCD5331",
		"27AC4454504D33B9CC230DD502202BA988D0CC84AE20034200EDA82152B05D5E81099969A21801D2",
		"4D020860CB1F40408B1DB1243D468D11CCB14D1E4CE471873423689B64222CEE54D33290102DE080",
		"BF1D62E32BB22930C9C29743E0B17A23049C852D743CCFCC0CC7384189EF21603C3D494D3B0EA944",
		"0308A4460701030443CE5A4A8BBD755D8AC84C3CC836D4BC5A3C0D54102F120A9C50C10670853102",
		"071CB430F1BBE9B06566288AE4A52F5FE4DC1686C358380B05002884BFD829297023C4087443880A",
		"B0831ED7BBCDC882C2002BFC25051680010C6EE30F8F018710088449E90618AA22C4494EE6E0C25B",
		"5AA0000910B0844290420B1A302B6D3C00030280E11106A1089BD890363A3C82261EF080A5059109",
		"2F3CA21297C8C42606010021FF0B4D414347436F6E2004031039000000015772697474656E206279",
		"20474946436F6E76657274657220322E342E33206F66204D6F6E6461792C204D61792032352C2031",
		"393938003B"
	};

	private static final String[][] TURTLE_IMAGE = {
		TURTLE_00,
		TURTLE_01,
		TURTLE_02,
		TURTLE_03,
		TURTLE_04,
		TURTLE_05,
		TURTLE_06,
		TURTLE_07,
		TURTLE_08,
		TURTLE_09,
		TURTLE_10,
		TURTLE_11,
		TURTLE_12,
		TURTLE_13,
		TURTLE_14,
		TURTLE_15,
		TURTLE_16,
		TURTLE_17,
		TURTLE_18,
		TURTLE_19,
		TURTLE_20,
		TURTLE_21,
		TURTLE_22,
		TURTLE_23,
		TURTLE_24,
		TURTLE_25,
		TURTLE_26,
		TURTLE_27,
		TURTLE_28,
		TURTLE_29,
		TURTLE_30,
		TURTLE_31,
		TURTLE_32,
		TURTLE_33,
		TURTLE_34,
		TURTLE_35
	};
}
